#!/usr/bin/env python
"""
A script to demonstrate how to use your own source model
"""
import bilby
import numpy as np

# First set up logging and some output directories and labels
outdir = "outdir"
label = "create_your_own_source_model"
sampling_frequency = 4096
duration = 1


# Here we define out source model - this is the sine-Gaussian model in the
# frequency domain.
def gaussian(frequency_array, amplitude, f0, tau, phi0):
    r"""
    Our custom source model, this is just a Gaussian in frequency with
    variable global phase.

    .. math::

        \tilde{h}_{\plus}(f) = \frac{A \tau}{2\sqrt{\pi}}}
        e^{- \pi \tau (f - f_{0})^2 + i \phi_{0}} \\
        \tilde{h}_{\times}(f) = \tilde{h}_{\plus}(f) e^{i \pi / 2}


    Parameters
    ----------
    frequency_array: array-like
        The frequencies to evaluate the model at. This is required for all
        Bilby source models.
    amplitude: float
        An overall amplitude prefactor.
    f0: float
        The central frequency.
    tau: float
        The damping rate.
    phi0: float
        The reference phase.

    Returns
    -------
    dict:
        A dictionary containing "plus" and "cross" entries.

    """
    arg = -((np.pi * tau * (frequency_array - f0)) ** 2) + 1j * phi0
    plus = np.sqrt(np.pi) * amplitude * tau * np.exp(arg) / 2.0
    cross = plus * np.exp(1j * np.pi / 2)
    return {"plus": plus, "cross": cross}


# We now define some parameters that we will inject
injection_parameters = dict(
    amplitude=1e-23, f0=100, tau=1, phi0=0, geocent_time=0, ra=0, dec=0, psi=0
)

# Now we pass our source function to the WaveformGenerator
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=gaussian,
)

# Set up interferometers.
ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 0.5,
)
ifos.inject_signal(
    waveform_generator=waveform_generator,
    parameters=injection_parameters,
    raise_error=False,
)

# Here we define the priors for the search. We use the injection parameters
# except for the amplitude, f0, and geocent_time
prior = injection_parameters.copy()
prior["amplitude"] = bilby.core.prior.LogUniform(
    minimum=1e-25, maximum=1e-21, latex_label="$\\mathcal{A}$"
)
prior["f0"] = bilby.core.prior.Uniform(90, 110, latex_label="$f_{0}$")

likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)

result = bilby.core.sampler.run_sampler(
    likelihood,
    prior,
    sampler="dynesty",
    outdir=outdir,
    label=label,
    resume=False,
    sample="unif",
    injection_parameters=injection_parameters,
)
result.plot_corner()



#!/usr/bin/env python
"""
Tutorial to demonstrate running parameter estimation on a reduced parameter
space for an injected signal.

This example estimates the masses using a uniform prior in both component masses
and distance using a uniform in comoving volume prior on luminosity distance
between luminosity distances of 100Mpc and 5Gpc, the cosmology is Planck15.
"""

import bilby
import numpy as np

# Set the duration and sampling frequency of the data segment that we're
# going to inject the signal into
duration = 4.0
sampling_frequency = 2048.0
minimum_frequency = 20

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "fast_tutorial"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(88170235)

# We are going to inject a binary black hole waveform.  We first establish a
# dictionary of parameters that includes all of the different waveform
# parameters, including masses of the two black holes (mass_1, mass_2),
# spins of both black holes (a, tilt, phi), etc.
injection_parameters = dict(
    mass_1=36.0,
    mass_2=29.0,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=2000.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)

# Fixed arguments passed into the source model
waveform_arguments = dict(
    waveform_approximant="IMRPhenomPv2",
    reference_frequency=50.0,
    minimum_frequency=minimum_frequency,
)

# Create the waveform_generator using a LAL BinaryBlackHole source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
    waveform_arguments=waveform_arguments,
)

# Set up interferometers.  In this case we'll use two interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1). These default to their design
# sensitivity
ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 2,
)
ifos.inject_signal(
    waveform_generator=waveform_generator, parameters=injection_parameters
)

# Set up a PriorDict, which inherits from dict.
# By default we will sample all terms in the signal models.  However, this will
# take a long time for the calculation, so for this example we will set almost
# all of the priors to be equall to their injected values.  This implies the
# prior is a delta function at the true, injected value.  In reality, the
# sampler implementation is smart enough to not sample any parameter that has
# a delta-function prior.
# The above list does *not* include mass_1, mass_2, theta_jn and luminosity
# distance, which means those are the parameters that will be included in the
# sampler.  If we do nothing, then the default priors get used.
priors = bilby.gw.prior.BBHPriorDict()
for key in [
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "phi_12",
    "phi_jl",
    "psi",
    "ra",
    "dec",
    "geocent_time",
    "phase",
]:
    priors[key] = injection_parameters[key]

# Perform a check that the prior does not extend to a parameter space longer than the data
priors.validate_prior(duration, minimum_frequency)

# Initialise the likelihood by passing in the interferometer data (ifos) and
# the waveform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)

# Run sampler.  In this case we're going to use the `dynesty` sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    npoints=1000,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

# Make a corner plot.
result.plot_corner()



import itertools
import os
import pytest
import unittest
from copy import deepcopy
from itertools import product
from parameterized import parameterized

import numpy as np
import bilby
from bilby.gw.detector import calibration
from scipy.special import logsumexp


class TestMarginalizedLikelihood(unittest.TestCase):
    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=4000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
        )

        self.interferometers = bilby.gw.detector.InterferometerList(["H1"])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=self.parameters["geocent_time"] - self.duration / 2,
        )

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.prior = bilby.gw.prior.BBHPriorDict()
        self.prior["geocent_time"] = bilby.prior.Uniform(
            minimum=self.parameters["geocent_time"] - self.duration / 2,
            maximum=self.parameters["geocent_time"] + self.duration / 2,
        )

    def test_cannot_instantiate_marginalised_likelihood_without_prior(self):
        self.assertRaises(
            ValueError,
            lambda: bilby.gw.likelihood.GravitationalWaveTransient(
                interferometers=self.interferometers,
                waveform_generator=self.waveform_generator,
                phase_marginalization=True,
            ),
        )

    def test_generating_default_time_prior(self):
        temp = self.prior.pop("geocent_time")
        new_prior = self.prior.copy()
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=new_prior,
            time_marginalization=True,
        )
        same = all(
            [
                temp.minimum == like.priors["geocent_time"].minimum,
                temp.maximum == like.priors["geocent_time"].maximum,
                new_prior["geocent_time"] == temp.minimum,
            ]
        )
        self.assertTrue(same)
        self.prior["geocent_time"] = temp

    def test_generating_default_phase_prior(self):
        temp = self.prior.pop("phase")
        new_prior = self.prior.copy()
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=new_prior,
            phase_marginalization=True,
        )
        same = all(
            [
                temp.minimum == like.priors["phase"].minimum,
                temp.maximum == like.priors["phase"].maximum,
                new_prior["phase"] == float(0),
            ]
        )
        self.assertTrue(same)
        self.prior["phase"] = temp

    def test_run_sampler_flags_if_marginalized_phase_is_sampled(self):
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior,
            phase_marginalization=True,
        )
        new_prior = self.prior.copy()
        new_prior["phase"] = bilby.prior.Uniform(minimum=0, maximum=2 * np.pi)
        for key, param in dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            theta_jn=0.4,
            psi=2.659,
            ra=1.375,
            dec=-1.2108,
        ).items():
            new_prior[key] = param
        with self.assertRaises(bilby.core.sampler.SamplingMarginalisedParameterError):
            bilby.run_sampler(like, new_prior)

    def test_run_sampler_flags_if_marginalized_time_is_sampled(self):
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior,
            time_marginalization=True,
        )
        new_prior = self.prior.copy()
        new_prior["geocent_time"] = bilby.prior.Uniform(minimum=0, maximum=1)
        for key, param in dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            theta_jn=0.4,
            psi=2.659,
            ra=1.375,
            dec=-1.2108,
        ).items():
            new_prior[key] = param
        with self.assertRaises(bilby.core.sampler.SamplingMarginalisedParameterError):
            bilby.run_sampler(like, new_prior)


class TestMarginalizations(unittest.TestCase):
    """
    Test all marginalised likelihoods matches brute force version.

    For time, this is strongly dependent on the specific time grid used.
    The `time_jitter` parameter makes this a weaker dependence during sampling.
    """
    _parameters = product(
        ["regular", "roq", "relbin"],
        ["luminosity_distance", "geocent_time", "phase"],
        [True, False],
        [True, False],
        [True, False],
    )

    lookup_phase = "distance_lookup_phase.npz"
    lookup_no_phase = "distance_lookup_no_phase.npz"
    path_to_roq_weights = "weights.npz"

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=4000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
            time_jitter=0,
            fiducial=0,
        )

        self.interferometers = bilby.gw.detector.InterferometerList(["H1"])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=1126259640,
        )

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            start_time=1126259640,
            waveform_arguments=dict(
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
            )
        )
        self.interferometers.inject_signal(
            parameters=self.parameters, waveform_generator=self.waveform_generator
        )

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors["geocent_time"] = bilby.prior.Uniform(
            minimum=self.parameters["geocent_time"] - 0.1,
            maximum=self.parameters["geocent_time"] + 0.1
        )

        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), "ROQ_data/IMRPhenomPv2/4s"),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s",
        ]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        self.roq_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            start_time=1126259640,
            waveform_arguments=dict(
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
                frequency_nodes_linear=np.load(f"{roq_dir}/fnodes_linear.npy"),
                frequency_nodes_quadratic=np.load(f"{roq_dir}/fnodes_quadratic.npy"),
            )
        )
        self.roq_linear_matrix_file = f"{roq_dir}/B_linear.npy"
        self.roq_quadratic_matrix_file = f"{roq_dir}/B_quadratic.npy"

        self.relbin_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole_relative_binning,
            start_time=1126259640,
            waveform_arguments=dict(
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
            )
        )

    def tearDown(self):
        del self.duration
        del self.sampling_frequency
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.roq_waveform_generator
        del self.priors

    @classmethod
    def tearDownClass(cls):
        """remove lookup tables so that they are not used accidentally in subsequent tests"""
        for filename in [cls.lookup_phase, cls.lookup_no_phase, cls.path_to_roq_weights]:
            if os.path.exists(filename):
                os.remove(filename)

    def likelihood_kwargs(self, kind, time_marginalization, phase_marginalization, distance_marginalization, priors):
        if priors is None:
            priors = deepcopy(self.priors)
        if distance_marginalization and phase_marginalization:
            lookup = TestMarginalizations.lookup_phase
        elif distance_marginalization:
            lookup = TestMarginalizations.lookup_no_phase
        else:
            lookup = None
        kwargs = dict(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            distance_marginalization=distance_marginalization,
            phase_marginalization=phase_marginalization,
            time_marginalization=time_marginalization,
            distance_marginalization_lookup_table=lookup,
            priors=priors,
        )
        if kind == "roq":
            kwargs.update(dict(
                linear_matrix=self.roq_linear_matrix_file,
                quadratic_matrix=self.roq_quadratic_matrix_file,
                waveform_generator=self.roq_waveform_generator,
            ))
            if os.path.exists(self.__class__.path_to_roq_weights):
                kwargs["weights"] = self.__class__.path_to_roq_weights
        elif kind == "relbin":
            kwargs["fiducial_parameters"] = deepcopy(self.parameters)
            kwargs["waveform_generator"] = self.relbin_waveform_generator
        return kwargs

    def get_likelihood(
        self,
        kind,
        time_marginalization=False,
        phase_marginalization=False,
        distance_marginalization=False,
        priors=None
    ):
        kwargs = self.likelihood_kwargs(
            kind, time_marginalization, phase_marginalization, distance_marginalization, priors
        )
        if kind == "regular":
            cls_ = bilby.gw.likelihood.GravitationalWaveTransient
        elif kind == "roq":
            cls_ = bilby.gw.likelihood.ROQGravitationalWaveTransient
        elif kind == "relbin":
            cls_ = bilby.gw.likelihood.RelativeBinningGravitationalWaveTransient
            kwargs["epsilon"] = 0.3
            self.parameters["fiducial"] = 0
        else:
            raise ValueError(f"kind {kind} not understood")
        like = cls_(**kwargs)
        if kind == "roq" and not os.path.exists(self.__class__.path_to_roq_weights):
            like.save_weights(self.__class__.path_to_roq_weights)
        like.parameters = self.parameters.copy()
        if time_marginalization:
            like.parameters["geocent_time"] = self.interferometers.start_time
        if distance_marginalization:
            like.parameters["luminosity_distance"] = like._ref_dist
        if phase_marginalization:
            like.parameters["phase"] = 0.0
        return like

    def _template(self, marginalized, non_marginalized, key, prior=None, values=None):
        if prior is None:
            prior = self.priors[key]
        if values is None:
            values = np.linspace(prior.minimum, prior.maximum, 1000)
        prior_values = prior.prob(values)
        ln_likes = np.empty(values.shape)
        for ii, value in enumerate(values):
            non_marginalized.parameters[key] = value
            ln_likes[ii] = non_marginalized.log_likelihood_ratio()
        like = np.exp(ln_likes - max(ln_likes))

        marg_like = np.log(np.trapz(like * prior_values, values)) + max(ln_likes)
        self.assertAlmostEqual(
            marg_like, marginalized.log_likelihood_ratio(), delta=0.5
        )

    @parameterized.expand(_parameters)
    def test_marginalisation(self, kind, key, distance, time, phase):
        if all([distance, time, phase]):
            pytest.skip()
        tested_args = dict(
            distance_marginalization=distance,
            time_marginalization=time,
            phase_marginalization=phase,
        )
        marg_key = f"{key.split('_')[-1]}_marginalization"
        if tested_args[marg_key]:
            pytest.skip()
        reference_args = tested_args.copy()
        reference_args[marg_key] = True
        self._template(
            self.get_likelihood(kind, **reference_args),
            self.get_likelihood(kind, **tested_args),
            key=key,
        )

    @parameterized.expand(["regular", "relbin"])
    def test_time_marginalisation_full_segment(self, kind):
        """
        Test time marginalised likelihood matches brute force version over
        just part of a segment.
        """
        priors = self.priors.copy()
        prior = bilby.prior.Uniform(
            minimum=self.interferometers.start_time,
            maximum=self.interferometers.start_time + self.interferometers.duration,
        )
        priors["geocent_time"] = prior
        self._template(
            self.get_likelihood(kind, time_marginalization=True, priors=priors.copy()),
            self.get_likelihood(kind, priors=priors.copy()),
            key="geocent_time",
            values=self.waveform_generator.time_array,
            prior=prior,
        )

    @parameterized.expand(
        itertools.product(["regular", "roq", "relbin"], *itertools.repeat([True, False], 3)),
        name_func=lambda func, num, param: (
            f"{func.__name__}_{num}__{param.args[0]}_" + "_".join([
                ["D", "P", "T"][ii] for ii, val
                in enumerate(param.args[1:]) if val
            ])
        )
    )
    def test_marginalization_reconstruction(self, kind, distance, phase, time):
        marginalizations = dict(
            geocent_time=time,
            luminosity_distance=distance,
            phase=phase,
        )
        like = self.get_likelihood(
            kind=kind,
            distance_marginalization=distance,
            time_marginalization=time,
            phase_marginalization=phase,
        )
        params = self.parameters.copy()
        reference_values = dict(
            luminosity_distance=self.priors["luminosity_distance"].rescale(0.5),
            geocent_time=self.interferometers.start_time,
            phase=0.0,
        )
        for key in marginalizations:
            if marginalizations[key]:
                params[key] = reference_values[key]
        like.parameters.update(params)
        output = like.generate_posterior_sample_from_marginalized_likelihood()
        for key in marginalizations:
            self.assertFalse(marginalizations[key] and reference_values[key] == output[key])


class CalibrationMarginalization(unittest.TestCase):

    def setUp(self):
        self.ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
        self.ifos.set_strain_data_from_power_spectral_densities(
            duration=4, sampling_frequency=1024
        )
        self.ifos[0].calibration_model = calibration.CubicSpline(
            prefix="recalib_H1_",
            minimum_frequency=20,
            maximum_frequency=512,
            n_points=5,
        )
        self.ifos[1].calibration_model = calibration.Precomputed.constant_uncertainty_spline(
            amplitude_sigma=0.1,
            phase_sigma=0.1,
            frequency_array=self.ifos[1].frequency_array[self.ifos[1].frequency_mask],
            n_nodes=5,
            label="L1",
            n_curves=1000,
        )
        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors["geocent_time"] = bilby.core.prior.Uniform(0, 4)
        self.priors.update(bilby.gw.prior.CalibrationPriorDict.constant_uncertainty_spline(
            amplitude_sigma=0.1,
            phase_sigma=0.1,
            minimum_frequency=20,
            maximum_frequency=512,
            n_nodes=5,
            label="H1",
        ))
        self.wfg = bilby.gw.waveform_generator.WaveformGenerator(
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
        )

    def tearDown(self):
        for ifo in self.ifos:
            filename = f"{ifo.name}_calibration_file.h5"
            if os.path.exists(filename):
                os.remove(filename)

    @parameterized.expand(["no_time", "time"])
    def test_calibration_marginalization(self, time):
        marginalized = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=deepcopy(self.ifos),
            waveform_generator=self.wfg,
            priors=deepcopy(self.priors),
            calibration_marginalization=True,
            time_marginalization=time == "time",
            number_of_response_curves=100,
            jitter_time=False,
        )
        non_marginalized = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=deepcopy(self.ifos),
            waveform_generator=self.wfg,
            priors=deepcopy(self.priors),
            calibration_marginalization=False,
            time_marginalization=time == "time",
            jitter_time=False,
        )
        parameters = self.priors.sample()
        marginalized.parameters.update(parameters)
        non_marginalized.parameters.update(parameters)
        marg_ln_l = marginalized.log_likelihood_ratio()
        non_marg_ln_ls = list()
        draws = marginalized.calibration_parameter_draws
        for ii in range(100):
            for name, value in draws["H1"].items():
                non_marginalized.parameters[name] = value[ii]
            non_marginalized.parameters["recalib_index_L1"] = draws["L1"]["recalib_index_L1"][ii]
            non_marg_ln_ls.append(non_marginalized.log_likelihood_ratio())
        non_marg_ln_l = logsumexp(non_marg_ln_ls, b=1 / 100)
        self.assertAlmostEqual(marg_ln_l, non_marg_ln_l)



#!/usr/bin/env python
"""
Demonstrate using the FakeSampler to reweight a result to include higher-order
emission modes. This is a simplified version of the method presented in
arXiv:1905.05477, however, the method can be applied to a much wider range of
initial and more complex likelihoods.
"""

import bilby
import matplotlib.pyplot as plt
import numpy as np


def make_comparison_histograms(result_1, result_2):
    pos_full = result_1.posterior
    pos_simple = result_2.posterior

    plt.figure()
    plt.hist(
        pos_full["log_likelihood"],
        50,
        label=result_1.label,
        histtype="step",
        density=True,
    )
    plt.hist(
        pos_simple["log_likelihood"],
        50,
        label=result_2.label,
        histtype="step",
        density=True,
    )
    plt.xlabel(r"delta_logl")
    plt.legend(loc=2)
    plt.savefig(f"{result_1.outdir}/delta_logl.pdf")
    plt.close()


def main():
    outdir = "outdir"

    np.random.seed(170808)

    duration = 4
    sampling_frequency = 1024

    injection_parameters = dict(
        chirp_mass=36.0,
        mass_ratio=0.2,
        chi_1=0.4,
        chi_2=0.3,
        luminosity_distance=2000.0,
        theta_jn=0.4,
        psi=0.659,
        phase=1.3,
        geocent_time=1126259642.413,
        ra=1.375,
        dec=-1.2108,
    )

    waveform_arguments = dict(
        waveform_approximant="IMRPhenomXAS",
        reference_frequency=20.0,
        minimum_frequency=20.0,
    )

    waveform_generator = bilby.gw.WaveformGenerator(
        duration=duration,
        sampling_frequency=sampling_frequency,
        frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        waveform_arguments=waveform_arguments,
        parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
    )

    ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])

    ifos.set_strain_data_from_zero_noise(
        sampling_frequency=sampling_frequency,
        duration=duration,
        start_time=injection_parameters["geocent_time"] - 2,
    )

    ifos.inject_signal(
        waveform_generator=waveform_generator,
        parameters=injection_parameters,
    )

    priors = bilby.gw.prior.BBHPriorDict(aligned_spin=True)
    for key in [
        "luminosity_distance",
        "theta_jn",
        "phase",
        "psi",
        "ra",
        "dec",
        "geocent_time",
    ]:
        priors[key] = injection_parameters[key]
    priors["chirp_mass"].maximum = 45

    likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
        interferometers=ifos, waveform_generator=waveform_generator
    )

    # perform the initial sampling with our simple model
    original_result = bilby.run_sampler(
        likelihood=likelihood,
        priors=priors,
        sampler="dynesty",
        walks=10,
        nact=3,
        bound="single",
        injection_parameters=injection_parameters,
        outdir=outdir,
        label="primary_mode_only",
        save="hdf5",
    )

    # update the waveform generator to use our higher-order mode waveform
    likelihood.waveform_generator.waveform_arguments[
        "waveform_approximant"
    ] = "IMRPhenomXHM"

    # call the FakeSampler to compute the new likelihoods
    new_result = bilby.run_sampler(
        likelihood=likelihood,
        priors=priors,
        sampler="fake_sampler",
        sample_file=f"{outdir}/{original_result.label}_result.hdf5",
        injection_parameters=injection_parameters,
        outdir=outdir,
        verbose=False,
        label="higher_order_mode",
        save="hdf5",
    )

    # make some comparison plots
    bilby.core.result.plot_multiple([original_result, new_result])
    make_comparison_histograms(new_result, original_result)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Tutorial to demonstrate running parameter estimation sampling in non-standard
parameters for an injected signal.

This example estimates the masses using a uniform prior in chirp mass,
mass ratio and redshift.
The cosmology is according to the Planck 2015 data release.
"""
import bilby
import numpy as np

bilby.core.utils.setup_logger(log_level="info")

duration = 4
sampling_frequency = 2048
outdir = "outdir"
label = "different_parameters"

np.random.seed(151226)

injection_parameters = dict(
    total_mass=66.0,
    mass_ratio=0.9,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=2000,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)

waveform_arguments = dict(waveform_approximant="IMRPhenomXP", reference_frequency=50.0)

# Create the waveform_generator using a LAL BinaryBlackHole source function
# We specify a function which transforms a dictionary of parameters into the
# appropriate parameters for the source model.
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    sampling_frequency=sampling_frequency,
    duration=duration,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
    waveform_arguments=waveform_arguments,
)

# Set up interferometers.
ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1", "K1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 2,
)
ifos.inject_signal(
    waveform_generator=waveform_generator, parameters=injection_parameters
)

# Set up prior
# Note it is possible to sample in different parameters to those that were
# injected.
priors = bilby.gw.prior.BBHPriorDict()

del priors["mass_ratio"]
priors["symmetric_mass_ratio"] = bilby.prior.Uniform(
    name="symmetric_mass_ratio", latex_label="q", minimum=0.1, maximum=0.25
)

del priors["luminosity_distance"]
priors["redshift"] = bilby.prior.Uniform(
    name="redshift", latex_label="$z$", minimum=0, maximum=0.5
)
# These parameters will not be sampled
for key in [
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "phi_12",
    "phi_jl",
    "psi",
    "ra",
    "dec",
    "geocent_time",
    "phase",
]:
    priors[key] = injection_parameters[key]
del priors["theta_jn"]
priors["cos_theta_jn"] = np.cos(injection_parameters["theta_jn"])
print(priors)

# Initialise GravitationalWaveTransient
likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)

# Run sampler
# Note we've added a post-processing conversion function, this will generate
# many useful additional parameters, e.g., source-frame masses.
result = bilby.core.sampler.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    walks=25,
    nact=5,
    outdir=outdir,
    injection_parameters=injection_parameters,
    label=label,
    conversion_function=bilby.gw.conversion.generate_all_bbh_parameters,
)
result.plot_corner()



#!/usr/bin/env python
"""
An example of how to use bilby to perform parameter estimation for
non-gravitational wave data. In this case, fitting a linear function to
data with background Gaussian noise

"""
import bilby
import matplotlib.pyplot as plt
import numpy as np

# A few simple setup steps
label = "linear_regression_bilby_mcmc"
outdir = "outdir"
bilby.utils.check_directory_exists_and_if_not_mkdir(outdir)


# First, we define our "signal model", in this case a simple linear function
def model(time, m, c):
    return time * m + c


# Now we define the injection parameters which we make simulated data with
injection_parameters = dict(m=0.5, c=0.2)

# For this example, we'll use standard Gaussian noise

# These lines of code generate the fake data. Note the ** just unpacks the
# contents of the injection_parameters when calling the model function.
sampling_frequency = 10
time_duration = 10
time = np.arange(0, time_duration, 1 / sampling_frequency)
N = len(time)
sigma = np.random.normal(1, 0.01, N)
data = model(time, **injection_parameters) + np.random.normal(0, sigma, N)

# We quickly plot the data to check it looks sensible
fig, ax = plt.subplots()
ax.plot(time, data, "o", label="data")
ax.plot(time, model(time, **injection_parameters), "--r", label="signal")
ax.set_xlabel("time")
ax.set_ylabel("y")
ax.legend()
fig.savefig("{}/{}_data.png".format(outdir, label))

# Now lets instantiate a version of our GaussianLikelihood, giving it
# the time, data and signal model
likelihood = bilby.likelihood.GaussianLikelihood(time, data, model, sigma)

# From hereon, the syntax is exactly equivalent to other bilby examples
# We make a prior
priors = dict()
priors["m"] = bilby.core.prior.Uniform(0, 5, "m")
priors["c"] = bilby.core.prior.Uniform(-2, 2, "c")

# And run sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="bilby_mcmc",
    nsamples=10000,
    ntemps=1,
    nensemble=1,
    L1steps=1,
    print_dt=5,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

# Finally plot a corner plot: all outputs are stored in outdir
result.plot_corner()



#!/usr/bin/env python
"""
Tutorial to demonstrate running parameter estimation with calibration
uncertainties marginalized over using a finite set of realizations.
"""

from copy import deepcopy

import bilby
import matplotlib.pyplot as plt
import numpy as np

# Set the duration and sampling frequency of the data segment
# that we're going to create and inject the signal into.
duration = 4
sampling_frequency = 1024

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "calibration_marginalization"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(170817)

# We are going to inject a binary black hole waveform. We first establish a
# dictionary of parameters that includes all of the different waveform
# parameters, including masses of the two black holes (mass_1, mass_2),
# spins of both black holes (a, tilt, phi), etc.
injection_parameters = dict(
    mass_1=36.0,
    mass_2=29.0,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=2000.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)
start_time = injection_parameters["geocent_time"] - duration + 2

# Fixed arguments passed into the source model
waveform_arguments = dict(waveform_approximant="IMRPhenomXP", reference_frequency=50.0)

# Create the waveform_generator using a LAL BinaryBlackHole source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameters=injection_parameters,
    waveform_arguments=waveform_arguments,
)
waveform_generator_rew = deepcopy(waveform_generator)

# Set up interferometers. In this case we'll use three interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1), and Virgo (V1)).
# These default to their design sensitivity
ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
for ifo in ifos:
    injection_parameters.update(
        {f"recalib_{ifo.name}_amplitude_{ii}": 0.0 for ii in range(10)}
    )
    injection_parameters.update(
        {f"recalib_{ifo.name}_phase_{ii}": 0.0 for ii in range(10)}
    )
    ifo.calibration_model = bilby.gw.calibration.CubicSpline(
        prefix=f"recalib_{ifo.name}_",
        minimum_frequency=ifo.minimum_frequency,
        maximum_frequency=ifo.maximum_frequency,
        n_points=10,
    )
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency, duration=duration, start_time=start_time
)
ifos.inject_signal(
    parameters=injection_parameters, waveform_generator=waveform_generator
)
ifos_rew = deepcopy(ifos)

# Set up prior, which is a dictionary
# Here we fix the injected cbc parameters (except the distance)
# to the injected values.
priors = injection_parameters.copy()
priors["luminosity_distance"] = bilby.prior.Uniform(
    injection_parameters["luminosity_distance"] - 1000,
    injection_parameters["luminosity_distance"] + 1000,
    name="luminosity_distance",
    latex_label="$d_L$",
)
for key in injection_parameters:
    if "recalib" in key:
        priors[key] = injection_parameters[key]

# Convert to prior dictionary to replace the floats with delta function priors
priors = bilby.core.prior.PriorDict(priors)
priors_rew = deepcopy(priors)

# Initialise the likelihood by passing in the interferometer data (IFOs) and
# the waveform generator. Here we assume there is no calibration uncertainty
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator, priors=priors
)

# Run sampler.  In this case we're going to use the `dynesty` sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    npoints=500,
    walks=20,
    nact=3,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

# Setting the log likelihood to actually be the log likelihood and not the log likelihood ratio...
# This is used the for reweighting
result.posterior["log_likelihood"] = (
    result.posterior["log_likelihood"] + result.log_noise_evidence
)

# Setting the priors we want on the calibration response curve parameters - as an example.
for name in ["recalib_H1_amplitude_1", "recalib_H1_amplitude_4"]:
    priors_rew[name] = bilby.prior.Gaussian(
        mu=0, sigma=0.03, name=name, latex_label=f"H1 $A_{name[-1]}$"
    )

# Setting up the calibration marginalized likelihood.
# We save the calibration response curve files into the output directory under {ifo.name}_calibration_file.h5
cal_likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos_rew,
    waveform_generator=waveform_generator_rew,
    calibration_marginalization=True,
    priors=priors_rew,
    number_of_response_curves=100,
    calibration_lookup_table={
        ifos[i].name: f"{outdir}/{ifos[i].name}_calibration_file.h5"
        for i in range(len(ifos))
    },
)

# Plot the magnitude of the curves to be used in the marginalization
plt.semilogx(
    ifos[0].frequency_array[ifos[0].frequency_mask][0:-1],
    cal_likelihood.calibration_draws[ifos[0].name][:, 0:-1].T,
)
plt.xlim(20, 1024)
plt.ylabel("Magnitude")
plt.xlabel("Frequency [Hz]")
plt.savefig(f"{outdir}/calibration_draws.pdf")
plt.clf()

# Reweight the posterior samples from a distribution with no calibration uncertainty to one with uncertainty.
# This method utilizes rejection sampling which can be inefficient at drawing samples at higher SNRs.
result_rew = bilby.core.result.reweight(
    result,
    new_likelihood=cal_likelihood,
    conversion_function=bilby.gw.conversion.generate_all_bbh_parameters,
)

# Plot distance posterior with and without the calibration
for res, label in zip(
    [result, result_rew], ["No calibration uncertainty", "Calibration uncertainty"]
):
    plt.hist(
        res.posterior["luminosity_distance"],
        label=label,
        bins=50,
        histtype="step",
        density=True,
    )
plt.legend()
plt.xlabel("Luminosity distance [Mpc]")
plt.savefig(f"{outdir}/luminosity_distance_posterior.pdf")
plt.clf()

plt.hist(
    result_rew.posterior["recalib_index"],
    bins=np.linspace(
        0,
        cal_likelihood.number_of_response_curves - 1,
        cal_likelihood.number_of_response_curves,
    ),
    density=True,
)
plt.xlim(0, cal_likelihood.number_of_response_curves - 1)
plt.xlabel("Calibration index")
plt.savefig(f"{outdir}/calibration_index_histogram.pdf")



import json
import logging
from pathlib import Path
import subprocess
import sys

logger = logging.getLogger('bilby')


def setup_logger(outdir='.', label=None, log_level='INFO', print_version=False):
    """ Setup logging output: call at the start of the script to use

    Parameters
    ==========
    outdir, label: str
        If supplied, write the logging output to outdir/label.log
    log_level: str, optional
        ['debug', 'info', 'warning']
        Either a string from the list above, or an integer as specified
        in https://docs.python.org/2/library/logging.html#logging-levels
    print_version: bool
        If true, print version information
    """

    if type(log_level) is str:
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError('log_level {} not understood'.format(log_level))
    else:
        level = int(log_level)

    logger = logging.getLogger('bilby')
    logger.propagate = False
    logger.setLevel(level)

    if any([type(h) == logging.StreamHandler for h in logger.handlers]) is False:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if any([type(h) == logging.FileHandler for h in logger.handlers]) is False:
        if label:
            Path(outdir).mkdir(parents=True, exist_ok=True)
            log_file = '{}/{}.log'.format(outdir, label)
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    if print_version:
        version = get_version_information()
        logger.info('Running bilby version: {}'.format(version))


def get_version_information():
    from bilby import __version__
    return __version__


def loaded_modules_dict():
    module_names = list(sys.modules.keys())
    vdict = {}
    for key in module_names:
        if "." not in str(key):
            vdict[key] = str(getattr(sys.modules[key], "__version__", "N/A"))
    return vdict


def env_package_list(as_dataframe=False):
    """Get the list of packages installed in the system prefix.

    If it is detected that the system prefix is part of a Conda environment,
    a call to ``conda list --prefix {sys.prefix}`` will be made, otherwise
    the call will be to ``{sys.executable} -m pip list installed``.

    Parameters
    ----------
    as_dataframe: bool
        return output as a `pandas.DataFrame`

    Returns
    -------
    pkgs : `list` of `dict`, or `pandas.DataFrame`
        If ``as_dataframe=False`` is given, the output is a `list` of `dict`,
        one for each package, at least with ``'name'`` and ``'version'`` keys
        (more if `conda` is used).
        If ``as_dataframe=True`` is given, the output is a `DataFrame`
        created from the `list` of `dicts`.
    """
    prefix = sys.prefix

    # if a conda-meta directory exists, this is a conda environment, so
    # use conda to print the package list
    conda_detected = (Path(prefix) / "conda-meta").is_dir()
    if conda_detected:
        try:
            pkgs = json.loads(subprocess.check_output([
                "conda",
                "list",
                "--prefix", prefix,
                "--json"
            ]))
        except (FileNotFoundError, subprocess.CalledProcessError):
            # When a conda env is in use but conda is unavailable
            conda_detected = False

    # otherwise try and use Pip
    if not conda_detected:
        try:
            import pip  # noqa: F401
        except ModuleNotFoundError:  # no pip?
            # not a conda environment, and no pip, so just return
            # the list of loaded modules
            modules = loaded_modules_dict()
            pkgs = [{"name": x, "version": y} for x, y in modules.items()]
        else:
            pkgs = json.loads(subprocess.check_output([
                sys.executable,
                "-m", "pip",
                "list", "installed",
                "--format", "json",
            ]))

    # convert to recarray for storage
    if as_dataframe:
        from pandas import DataFrame
        return DataFrame(pkgs)
    return pkgs



#!/usr/bin/env python
"""
A script to demonstrate how to create and plot EoS's with Bilby
"""
from bilby.gw import eos

# In this script we're going to use Bilby to plot the MPA1 EoS from tabulated data
# and plot its representation as a spectral decomposition EoS.

# First, we specify the spectral parameter values for the MPA1 EoS.
MPA1_gammas = [1.0215, 0.1653, -0.0235, -0.0004]
MPA1_p0 = 1.51e33  # Pressure in CGS
MPA1_e0_c2 = 2.04e14  # \epsilon_0 / c^2 in CGS
MPA1_xmax = 6.63  # Dimensionless ending pressure

# Create the spectral decomposition EoS class
MPA1_spectral = eos.SpectralDecompositionEOS(
    MPA1_gammas, p0=MPA1_p0, e0=MPA1_e0_c2, xmax=MPA1_xmax, npts=100
)

# And create another from tabulated data
MPA1_tabulated = eos.TabularEOS("MPA1")

# Now let's plot them
# To do so, we specify a representation and plot ranges.
MPA1_spectral_plot = MPA1_spectral.plot(
    "pressure-energy_density", xlim=[1e22, 1e36], ylim=[1e9, 1e36]
)
MPA1_tabular_plot = MPA1_tabulated.plot(
    "pressure-energy_density", xlim=[1e22, 1e36], ylim=[1e9, 1e36]
)
MPA1_spectral_plot.savefig("spectral_mpa1.pdf")
MPA1_tabular_plot.savefig("tabular_mpa1.pdf")



#!/usr/bin/env python
"""
Tutorial to demonstrates reading and plotting the GraceDB data for GW170608,
the lightest GWTC-1 BBH event.

This tutorial must be run on a LIGO cluster.
"""

import bilby

label = "GW170608"
outdir = label + "_out"
# List of detectors involved in the event. GW170608 is pre-August 2017 so only
# H1 and L1 were involved
detectors = ["H1", "L1"]
# Names of the channels and query types to use, and the event GraceDB ID - ref.
# https://ldas-jobs.ligo.caltech.edu/~eve.chase/monitor/online_pe/C02_clean/
# C02_HL_Pv2/lalinferencenest/IMRPhenomPv2pseudoFourPN/config.ini
channel_names = ["H1:DCH-CLEAN_STRAIN_C02", "L1:DCH-CLEAN_STRAIN_C02"]
gracedb = "G288686"
# Duration of data around the event to use
duration = 16
# Duration of PSD data
psd_duration = 1024
# Minimum frequency and reference frequency
minimum_frequency = 10  # Hz
sampling_frequency = 2048.0

# Get trigger time
candidate = bilby.gw.utils.gracedb_to_json(gracedb, outdir=outdir)
trigger_time = candidate["gpstime"]
gps_start_time = trigger_time + 2.0 - duration

# Load the PSD data starting after the segment you want to analyze
psd_start_time = gps_start_time + duration

# Set up interferometer objects from the cache files
interferometers = bilby.gw.detector.InterferometerList([])

for channel_name in channel_names:
    interferometer = bilby.gw.detector.load_data_by_channel_name(
        start_time=gps_start_time,
        segment_duration=duration,
        psd_duration=psd_duration,
        psd_start_time=psd_start_time,
        channel_name=channel_name,
        sampling_frequency=sampling_frequency,
        outdir=outdir,
    )
    interferometer.minimum_frequency = minimum_frequency
    interferometers.append(interferometer)

# Plot the strain data
interferometers.plot_data(outdir=outdir, label=label)



#!/usr/bin/env python
"""
An example of how to use slab-and-spike priors in bilby.
In this example we look at a simple example with the sum
of two Gaussian distributions, and we try to fit with
up to three Gaussians.

We will use the `PyMultiNest` sampler which is fast but can be unreliable when
significant correlations exist in the likelihood.

To install `PyMultiNest` call

$ conda install -c conda-forge pymultinest
"""

import bilby
import matplotlib.pyplot as plt
import numpy as np

outdir = "outdir"
label = "slabspike"
bilby.utils.check_directory_exists_and_if_not_mkdir(outdir)


# Here we define our model. We want to inject two Gaussians and recover with up to three.
def gaussian(xs, amplitude, mu, sigma):
    return (
        amplitude
        / np.sqrt(2 * np.pi * sigma**2)
        * np.exp(-0.5 * (xs - mu) ** 2 / sigma**2)
    )


def triple_gaussian(
    xs,
    amplitude_0,
    amplitude_1,
    amplitude_2,
    mu_0,
    mu_1,
    mu_2,
    sigma_0,
    sigma_1,
    sigma_2,
    **kwargs,
):
    return (
        gaussian(xs, amplitude_0, mu_0, sigma_0)
        + gaussian(xs, amplitude_1, mu_1, sigma_1)
        + gaussian(xs, amplitude_2, mu_2, sigma_2)
    )


# Let's create our data set. We create 200 points on a grid.

xs = np.linspace(-5, 5, 200)
dx = xs[1] - xs[0]

# Note for our injection parameters we set the amplitude of the second component to 0.
injection_params = dict(
    amplitude_0=-3,
    mu_0=-4,
    sigma_0=4,
    amplitude_1=0,
    mu_1=0,
    sigma_1=1,
    amplitude_2=4,
    mu_2=3,
    sigma_2=3,
)

# We calculate the injected curve and add some Gaussian noise on the data points
sigma = 0.02
p = bilby.core.prior.Gaussian(mu=0, sigma=sigma)
ys = triple_gaussian(xs=xs, **injection_params) + p.sample(len(xs))

plt.errorbar(xs, ys, yerr=sigma, fmt=".k", capsize=0, label="Injected data")
plt.plot(xs, triple_gaussian(xs=xs, **injection_params), label="True signal")
plt.legend()
plt.savefig(f"{outdir}/{label}_injected_data")
plt.clf()


# Now we want to set up our priors.
priors = bilby.core.prior.PriorDict()
# For the slab-and-spike prior, we first need to define the 'slab' part, which is just a regular bilby prior.
amplitude_slab_0 = bilby.core.prior.Uniform(
    minimum=-10, maximum=10, name="amplitude_0", latex_label="$A_0$"
)
amplitude_slab_1 = bilby.core.prior.Uniform(
    minimum=-10, maximum=10, name="amplitude_1", latex_label="$A_1$"
)
amplitude_slab_2 = bilby.core.prior.Uniform(
    minimum=-10, maximum=10, name="amplitude_2", latex_label="$A_2$"
)
# We do the following to create the slab-and-spike prior. The spike height is somewhat arbitrary and can
# be corrected in post-processing.
priors["amplitude_0"] = bilby.core.prior.SlabSpikePrior(
    slab=amplitude_slab_0, spike_location=0, spike_height=0.1
)
priors["amplitude_1"] = bilby.core.prior.SlabSpikePrior(
    slab=amplitude_slab_1, spike_location=0, spike_height=0.1
)
priors["amplitude_2"] = bilby.core.prior.SlabSpikePrior(
    slab=amplitude_slab_2, spike_location=0, spike_height=0.1
)
# Our problem has a degeneracy in the ordering. In general, this problem is somewhat difficult to resolve properly.
# See e.g. https://github.com/GregoryAshton/kookaburra/blob/master/src/priors.py#L72 for an implementation.
# We resolve this by not letting the priors overlap in this case.
priors["mu_0"] = bilby.core.prior.Uniform(
    minimum=-5, maximum=-2, name="mu_0", latex_label=r"$\mu_0$"
)
priors["mu_1"] = bilby.core.prior.Uniform(
    minimum=-2, maximum=2, name="mu_1", latex_label=r"$\mu_1$"
)
priors["mu_2"] = bilby.core.prior.Uniform(
    minimum=2, maximum=5, name="mu_2", latex_label=r"$\mu_2$"
)
priors["sigma_0"] = bilby.core.prior.LogUniform(
    minimum=0.01, maximum=10, name="sigma_0", latex_label=r"$\sigma_0$"
)
priors["sigma_1"] = bilby.core.prior.LogUniform(
    minimum=0.01, maximum=10, name="sigma_1", latex_label=r"$\sigma_1$"
)
priors["sigma_2"] = bilby.core.prior.LogUniform(
    minimum=0.01, maximum=10, name="sigma_2", latex_label=r"$\sigma_2$"
)

# Setting up the likelihood and running the samplers works the same as elsewhere.
likelihood = bilby.core.likelihood.GaussianLikelihood(
    x=xs, y=ys, func=triple_gaussian, sigma=sigma
)
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    outdir=outdir,
    label=label,
    sampler="pymultinest",
    nlive=400,
)

result.plot_corner(truths=injection_params)


# Let's also plot the maximum likelihood fit along with the data.
max_like_params = result.posterior.iloc[-1]
plt.errorbar(xs, ys, yerr=sigma, fmt=".k", capsize=0, label="Injected data")
plt.plot(xs, triple_gaussian(xs=xs, **injection_params), label="True signal")
plt.plot(xs, triple_gaussian(xs=xs, **max_like_params), label="Max likelihood fit")
plt.legend()
plt.savefig(f"{outdir}/{label}_max_likelihood_recovery")
plt.clf()

# Finally, we can check what fraction of amplitude samples are exactly on the spike.
spike_samples_0 = len(np.where(result.posterior["amplitude_0"] == 0.0)[0]) / len(
    result.posterior
)
spike_samples_1 = len(np.where(result.posterior["amplitude_1"] == 0.0)[0]) / len(
    result.posterior
)
spike_samples_2 = len(np.where(result.posterior["amplitude_2"] == 0.0)[0]) / len(
    result.posterior
)
print(f"{spike_samples_0 * 100:.2f}% of amplitude_0 samples are exactly 0.0")
print(f"{spike_samples_1 * 100:.2f}% of amplitude_1 samples are exactly 0.0")
print(f"{spike_samples_2 * 100:.2f}% of amplitude_2 samples are exactly 0.0")



import datetime
import inspect
import os
import sys
import time

import numpy as np
from pandas import DataFrame

from ..result import rejection_sample
from ..utils import (
    check_directory_exists_and_if_not_mkdir,
    latex_plot_format,
    logger,
    safe_file_dump,
)
from .base_sampler import NestedSampler, Sampler, _SamplingContainer, signal_wrapper


def _set_sampling_kwargs(args):
    nact, maxmcmc, proposals, naccept = args
    _SamplingContainer.nact = nact
    _SamplingContainer.maxmcmc = maxmcmc
    _SamplingContainer.proposals = proposals
    _SamplingContainer.naccept = naccept


def _prior_transform_wrapper(theta):
    """Wrapper to the prior transformation. Needed for multiprocessing."""
    from .base_sampler import _sampling_convenience_dump

    return _sampling_convenience_dump.priors.rescale(
        _sampling_convenience_dump.search_parameter_keys, theta
    )


def _log_likelihood_wrapper(theta):
    """Wrapper to the log likelihood. Needed for multiprocessing."""
    from .base_sampler import _sampling_convenience_dump

    if _sampling_convenience_dump.priors.evaluate_constraints(
        {
            key: theta[ii]
            for ii, key in enumerate(_sampling_convenience_dump.search_parameter_keys)
        }
    ):
        params = {
            key: t
            for key, t in zip(_sampling_convenience_dump.search_parameter_keys, theta)
        }
        _sampling_convenience_dump.likelihood.parameters.update(params)
        if _sampling_convenience_dump.use_ratio:
            return _sampling_convenience_dump.likelihood.log_likelihood_ratio()
        else:
            return _sampling_convenience_dump.likelihood.log_likelihood()
    else:
        return np.nan_to_num(-np.inf)


class Dynesty(NestedSampler):
    """
    bilby wrapper of `dynesty.NestedSampler`
    (https://dynesty.readthedocs.io/en/latest/)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `dynesty.NestedSampler`, see
    documentation for that class for further help. Under Other Parameters below,
    we list commonly used kwargs and the Bilby defaults.

    Parameters
    ==========
    likelihood: likelihood.Likelihood
        A  object with a log_l method
    priors: bilby.core.prior.PriorDict, dict
        Priors to be used in the search.
        This has attributes for each parameter to be sampled.
    outdir: str, optional
        Name of the output directory
    label: str, optional
        Naming scheme of the output files
    use_ratio: bool, optional
        Switch to set whether or not you want to use the log-likelihood ratio
        or just the log-likelihood
    plot: bool, optional
        Switch to set whether or not you want to create traceplots
    skip_import_verification: bool
        Skips the check if the sampler is installed if true. This is
        only advisable for testing environments
    print_method: str ('tqdm')
        The method to use for printing. The options are:
        - 'tqdm': use a `tqdm` `pbar`, this is the default.
        - 'interval-$TIME': print to `stdout` every `$TIME` seconds,
          e.g., 'interval-10' prints every ten seconds, this does not print every iteration
        - else: print to `stdout` at every iteration
    exit_code: int
        The code which the same exits on if it hasn't finished sampling
    check_point: bool,
        If true, use check pointing.
    check_point_plot: bool,
        If true, generate a trace plot along with the check-point
    check_point_delta_t: float (600)
        The minimum checkpoint period (in seconds). Should the run be
        interrupted, it can be resumed from the last checkpoint.
    n_check_point: int, optional (None)
        The number of steps to take before checking whether to check_point.
    resume: bool
        If true, resume run from checkpoint (if available)
    maxmcmc: int (5000)
        The maximum length of the MCMC exploration to find a new point
    nact: int (2)
        The number of autocorrelation lengths for MCMC exploration.
        For use with the :code:`act-walk` and :code:`rwalk` sample methods.
        See the dynesty guide in the Bilby docs for more details.
    naccept: int (60)
        The expected number of accepted steps for MCMC exploration when using
        the :code:`acceptance-walk` sampling method.
    rejection_sample_posterior: bool (True)
        Whether to form the posterior by rejection sampling the nested samples.
        If False, the nested samples are resampled with repetition. This was
        the default behaviour in :code:`Bilby<=1.4.1` and leads to
        non-independent samples being produced.
    proposals: iterable (None)
        The proposal methods to use during MCMC. This can be some combination
        of :code:`"diff", "volumetric"`. See the dynesty guide in the Bilby docs
        for more details. default=:code:`["diff"]`.

    Other Parameters
    ================
    nlive: int, (1000)
        The number of live points, note this can also equivalently be given as
        one of [nlive, nlives, n_live_points, npoints]
    bound: {'live', 'live-multi', 'none', 'single', 'multi', 'balls', 'cubes'}, ('live')
        Method used to select new points
    sample: {'act-walk', 'acceptance-walk', 'unif', 'rwalk', 'slice',
             'rslice', 'hslice', 'rwalk_dynesty'}, ('act-walk')
        Method used to sample uniformly within the likelihood constraints,
        conditioned on the provided bounds
    walks: int (100)
        Number of walks taken if using the dynesty implemented sample methods
        Note that the default `walks` in dynesty itself is 25, although using
        `ndim * 10` can be a reasonable rule of thumb for new problems.
        For :code:`sample="act-walk"` and :code:`sample="rwalk"` this parameter
        has no impact on the sampling.
    dlogz: float, (0.1)
        Stopping criteria
    """

    @property
    def _dynesty_init_kwargs(self):
        params = inspect.signature(self.sampler_init).parameters
        kwargs = {
            key: param.default
            for key, param in params.items()
            if param.default != param.empty
        }
        kwargs["sample"] = "act-walk"
        kwargs["bound"] = "live"
        kwargs["update_interval"] = 600
        kwargs["facc"] = 0.2
        return kwargs

    @property
    def _dynesty_sampler_kwargs(self):
        params = inspect.signature(self.sampler_class.run_nested).parameters
        kwargs = {
            key: param.default
            for key, param in params.items()
            if param.default != param.empty
        }
        kwargs["save_bounds"] = False
        if "dlogz" in kwargs:
            kwargs["dlogz"] = 0.1
        return kwargs

    @property
    def default_kwargs(self):
        kwargs = self._dynesty_init_kwargs
        kwargs.update(self._dynesty_sampler_kwargs)
        return kwargs

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        skip_import_verification=False,
        check_point=True,
        check_point_plot=True,
        n_check_point=None,
        check_point_delta_t=600,
        resume=True,
        nestcheck=False,
        exit_code=130,
        print_method="tqdm",
        maxmcmc=5000,
        nact=2,
        naccept=60,
        rejection_sample_posterior=True,
        proposals=None,
        **kwargs,
    ):
        self.nact = nact
        self.naccept = naccept
        self.maxmcmc = maxmcmc
        self.proposals = proposals
        self.print_method = print_method
        self._translate_kwargs(kwargs)
        super(Dynesty, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            exit_code=exit_code,
            **kwargs,
        )
        self.n_check_point = n_check_point
        self.check_point = check_point
        self.check_point_plot = check_point_plot
        self.resume = resume
        self.rejection_sample_posterior = rejection_sample_posterior
        self._apply_dynesty_boundaries("periodic")
        self._apply_dynesty_boundaries("reflective")

        self.nestcheck = nestcheck

        if self.n_check_point is None:
            self.n_check_point = max(
                int(check_point_delta_t / self._log_likelihood_eval_time / 10), 10
            )
        self.check_point_delta_t = check_point_delta_t
        logger.info(f"Checkpoint every check_point_delta_t = {check_point_delta_t}s")

        self.resume_file = f"{self.outdir}/{self.label}_resume.pickle"
        self.sampling_time = datetime.timedelta()
        self.pbar = None

    @property
    def sampler_function_kwargs(self):
        return {key: self.kwargs[key] for key in self._dynesty_sampler_kwargs}

    @property
    def sampler_init_kwargs(self):
        return {key: self.kwargs[key] for key in self._dynesty_init_kwargs}

    def _translate_kwargs(self, kwargs):
        kwargs = super()._translate_kwargs(kwargs)
        if "nlive" not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["nlive"] = kwargs.pop(equiv)
        if "print_progress" not in kwargs:
            if "verbose" in kwargs:
                kwargs["print_progress"] = kwargs.pop("verbose")
        if "walks" not in kwargs:
            for equiv in self.walks_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["walks"] = kwargs.pop(equiv)
        if "queue_size" not in kwargs:
            for equiv in self.npool_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["queue_size"] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        if not self.kwargs["walks"]:
            self.kwargs["walks"] = 100
        if self.kwargs["print_func"] is None:
            self.kwargs["print_func"] = self._print_func
            if "interval" in self.print_method:
                self._last_print_time = datetime.datetime.now()
                self._print_interval = datetime.timedelta(
                    seconds=float(self.print_method.split("-")[1])
                )
        Sampler._verify_kwargs_against_default_kwargs(self)

    def _print_func(
        self,
        results,
        niter,
        ncall=None,
        dlogz=None,
        stop_val=None,
        nbatch=None,
        logl_min=-np.inf,
        logl_max=np.inf,
        *args,
        **kwargs,
    ):
        """Replacing status update for dynesty.result.print_func"""
        if "interval" in self.print_method:
            _time = datetime.datetime.now()
            if _time - self._last_print_time < self._print_interval:
                return
            else:
                self._last_print_time = _time

                # Add time in current run to overall sampling time
                total_time = self.sampling_time + _time - self.start_time

                # Remove fractional seconds
                total_time_str = str(total_time).split(".")[0]

        # Extract results at the current iteration.
        loglstar = results.loglstar
        delta_logz = results.delta_logz
        logz = results.logz
        logzvar = results.logzvar
        nc = results.nc
        bounditer = results.bounditer
        eff = results.eff

        # Adjusting outputs for printing.
        if delta_logz > 1e6:
            delta_logz = np.inf
        if 0.0 <= logzvar <= 1e6:
            logzerr = np.sqrt(logzvar)
        else:
            logzerr = np.nan
        if logz <= -1e6:
            logz = -np.inf
        if loglstar <= -1e6:
            loglstar = -np.inf

        if self.use_ratio:
            key = "logz-ratio"
        else:
            key = "logz"

        # Constructing output.
        string = list()
        string.append(f"bound:{bounditer:d}")
        string.append(f"nc:{nc:3d}")
        string.append(f"ncall:{ncall:.1e}")
        string.append(f"eff:{eff:0.1f}%")
        string.append(f"{key}={logz:0.2f}+/-{logzerr:0.2f}")
        if nbatch is not None:
            string.append(f"batch:{nbatch}")
        if logl_min > -np.inf:
            string.append(f"logl:{logl_min:.1f} < {loglstar:.1f} < {logl_max:.1f}")
        if dlogz is not None:
            string.append(f"dlogz:{delta_logz:0.3f}>{dlogz:0.2g}")
        else:
            string.append(f"stop:{stop_val:6.3f}")
        string = " ".join(string)

        if self.print_method == "tqdm":
            self.pbar.set_postfix_str(string, refresh=False)
            self.pbar.update(niter - self.pbar.n)
        else:
            print(f"{niter}it [{total_time_str} {string}]", file=sys.stdout, flush=True)

    def _apply_dynesty_boundaries(self, key):
        # The periodic kwargs passed into dynesty allows the parameters to
        # wander out of the bounds, this includes both periodic and reflective.
        # these are then handled in the prior_transform
        selected = list()
        for ii, param in enumerate(self.search_parameter_keys):
            if self.priors[param].boundary == key:
                logger.debug(f"Setting {key} boundary for {param}")
                selected.append(ii)
        if len(selected) == 0:
            selected = None
        self.kwargs[key] = selected

    def nestcheck_data(self, out_file):
        import nestcheck.data_processing

        ns_run = nestcheck.data_processing.process_dynesty_run(out_file)
        nestcheck_result = f"{self.outdir}/{self.label}_nestcheck.pickle"
        safe_file_dump(ns_run, nestcheck_result, "pickle")

    @property
    def nlive(self):
        return self.kwargs["nlive"]

    @property
    def sampler_init(self):
        from dynesty import NestedSampler

        return NestedSampler

    @property
    def sampler_class(self):
        from dynesty.sampler import Sampler

        return Sampler

    def _set_sampling_method(self):
        """
        Resolve the sampling method and sampler to use from the provided
        :code:`bound` and :code:`sample` arguments.

        This requires registering the :code:`bilby` specific methods in the
        appropriate locations within :code:`dynesty`.

        Additionally, some combinations of bound/sample/proposals are not
        compatible and so we either warn the user or raise an error.
        """
        import dynesty

        _set_sampling_kwargs((self.nact, self.maxmcmc, self.proposals, self.naccept))

        sample = self.kwargs["sample"]
        bound = self.kwargs["bound"]

        if sample not in ["rwalk", "act-walk", "acceptance-walk"] and bound in [
            "live",
            "live-multi",
        ]:
            logger.info(
                "Live-point based bound method requested with dynesty sample "
                f"'{sample}', overwriting to 'multi'"
            )
            self.kwargs["bound"] = "multi"
        elif bound == "live":
            from .dynesty_utils import LivePointSampler

            dynesty.dynamicsampler._SAMPLERS["live"] = LivePointSampler
        elif bound == "live-multi":
            from .dynesty_utils import MultiEllipsoidLivePointSampler

            dynesty.dynamicsampler._SAMPLERS[
                "live-multi"
            ] = MultiEllipsoidLivePointSampler
        elif sample == "acceptance-walk":
            raise DynestySetupError(
                "bound must be set to live or live-multi for sample=acceptance-walk"
            )
        elif self.proposals is None:
            logger.warning(
                "No proposals specified using dynesty sampling, defaulting "
                "to 'volumetric'."
            )
            self.proposals = ["volumetric"]
            _SamplingContainer.proposals = self.proposals
        elif "diff" in self.proposals:
            raise DynestySetupError(
                "bound must be set to live or live-multi to use differential "
                "evolution proposals"
            )

        if sample == "rwalk":
            logger.info(
                "Using the bilby-implemented rwalk sample method with ACT estimated walks. "
                f"An average of {2 * self.nact} steps will be accepted up to chain length "
                f"{self.maxmcmc}."
            )
            from .dynesty_utils import AcceptanceTrackingRWalk

            if self.kwargs["walks"] > self.maxmcmc:
                raise DynestySetupError("You have maxmcmc < walks (minimum mcmc)")
            if self.nact < 1:
                raise DynestySetupError("Unable to run with nact < 1")
            AcceptanceTrackingRWalk.old_act = None
            dynesty.nestedsamplers._SAMPLING["rwalk"] = AcceptanceTrackingRWalk()
        elif sample == "acceptance-walk":
            logger.info(
                "Using the bilby-implemented rwalk sampling with an average of "
                f"{self.naccept} accepted steps per MCMC and maximum length {self.maxmcmc}"
            )
            from .dynesty_utils import FixedRWalk

            dynesty.nestedsamplers._SAMPLING["acceptance-walk"] = FixedRWalk()
        elif sample == "act-walk":
            logger.info(
                "Using the bilby-implemented rwalk sampling tracking the "
                f"autocorrelation function and thinning by "
                f"{self.nact} with maximum length {self.nact * self.maxmcmc}"
            )
            from .dynesty_utils import ACTTrackingRWalk

            ACTTrackingRWalk._cache = list()
            dynesty.nestedsamplers._SAMPLING["act-walk"] = ACTTrackingRWalk()
        elif sample == "rwalk_dynesty":
            sample = sample.strip("_dynesty")
            self.kwargs["sample"] = sample
            logger.info(f"Using the dynesty-implemented {sample} sample method")

    @signal_wrapper
    def run_sampler(self):
        import dynesty

        logger.info(f"Using dynesty version {dynesty.__version__}")

        self._set_sampling_method()
        self._setup_pool()

        if self.resume:
            self.resume = self.read_saved_state(continuing=True)

        if self.resume:
            logger.info("Resume file successfully loaded.")
        else:
            if self.kwargs["live_points"] is None:
                self.kwargs["live_points"] = self.get_initial_points_from_prior(
                    self.nlive
                )
                self.kwargs["live_points"] = (*self.kwargs["live_points"], None)
            self.sampler = self.sampler_init(
                loglikelihood=_log_likelihood_wrapper,
                prior_transform=_prior_transform_wrapper,
                ndim=self.ndim,
                **self.sampler_init_kwargs,
            )
        if self.print_method == "tqdm" and self.kwargs["print_progress"]:
            from tqdm.auto import tqdm

            self.pbar = tqdm(file=sys.stdout, initial=self.sampler.it)

        self.start_time = datetime.datetime.now()
        if self.check_point:
            out = self._run_external_sampler_with_checkpointing()
        else:
            out = self._run_external_sampler_without_checkpointing()
        self._update_sampling_time()

        self._close_pool()

        # Flushes the output to force a line break
        if self.pbar is not None:
            self.pbar = self.pbar.close()
            print("")

        check_directory_exists_and_if_not_mkdir(self.outdir)

        if self.nestcheck:
            self.nestcheck_data(out)

        dynesty_result = f"{self.outdir}/{self.label}_dynesty.pickle"
        safe_file_dump(out, dynesty_result, "dill")

        self._generate_result(out)
        self.result.sampling_time = self.sampling_time

        return self.result

    def _setup_pool(self):
        """
        In addition to the usual steps, we need to set the sampling kwargs on
        every process. To make sure we get every process, run the kwarg setting
        more times than we have processes.
        """
        super(Dynesty, self)._setup_pool()
        if self.pool is not None:
            args = (
                [(self.nact, self.maxmcmc, self.proposals, self.naccept)]
                * self.npool
                * 10
            )
            self.pool.map(_set_sampling_kwargs, args)

    def _generate_result(self, out):
        """
        Extract the information we need from the dynesty output. This includes
        the evidence, nested samples, run statistics. In addition, we generate
        the posterior samples from the nested samples.

        Parameters
        ==========
        out: dynesty.result.Result
            The dynesty output.
        """
        import dynesty
        from scipy.special import logsumexp

        logwts = out["logwt"]
        weights = np.exp(logwts - out["logz"][-1])
        nested_samples = DataFrame(out.samples, columns=self.search_parameter_keys)
        nested_samples["weights"] = weights
        nested_samples["log_likelihood"] = out.logl
        self.result.nested_samples = nested_samples
        if self.rejection_sample_posterior:
            keep = weights > np.random.uniform(0, max(weights), len(weights))
            self.result.samples = out.samples[keep]
            self.result.log_likelihood_evaluations = out.logl[keep]
            logger.info(
                f"Rejection sampling nested samples to obtain {sum(keep)} posterior samples"
            )
        else:
            self.result.samples = dynesty.utils.resample_equal(out.samples, weights)
            self.result.log_likelihood_evaluations = self.reorder_loglikelihoods(
                unsorted_loglikelihoods=out.logl,
                unsorted_samples=out.samples,
                sorted_samples=self.result.samples,
            )
            logger.info("Resampling nested samples to posterior samples in place.")
        self.result.log_evidence = out.logz[-1]
        self.result.log_evidence_err = out.logzerr[-1]
        self.result.information_gain = out.information[-1]
        self.result.num_likelihood_evaluations = getattr(self.sampler, "ncall", 0)

        logneff = logsumexp(logwts) * 2 - logsumexp(logwts * 2)
        neffsamples = int(np.exp(logneff))
        self.result.meta_data["run_statistics"] = dict(
            nlikelihood=self.result.num_likelihood_evaluations,
            neffsamples=neffsamples,
            sampling_time_s=self.sampling_time.seconds,
            ncores=self.kwargs.get("queue_size", 1),
        )

    def _update_sampling_time(self):
        end_time = datetime.datetime.now()
        self.sampling_time += end_time - self.start_time
        self.start_time = end_time

    def _run_external_sampler_without_checkpointing(self):
        logger.debug("Running sampler without checkpointing")
        self.sampler.run_nested(**self.sampler_function_kwargs)
        return self.sampler.results

    def finalize_sampler_kwargs(self, sampler_kwargs):
        sampler_kwargs["maxcall"] = self.n_check_point
        sampler_kwargs["add_live"] = True

    def _run_external_sampler_with_checkpointing(self):
        """
        In order to access the checkpointing, we run the sampler for short
        periods of time (less than the checkpoint time) and if sufficient
        time has passed, write a checkpoint before continuing. To get the most
        informative checkpoint plots, the current live points are added to the
        chain of nested samples within dynesty and have to be removed before
        restarting the sampler.
        """
        logger.debug("Running sampler with checkpointing")

        old_ncall = self.sampler.ncall
        sampler_kwargs = self.sampler_function_kwargs.copy()
        while True:
            self.finalize_sampler_kwargs(sampler_kwargs)
            self.sampler.run_nested(**sampler_kwargs)
            if self.sampler.ncall == old_ncall:
                break
            old_ncall = self.sampler.ncall

            if os.path.isfile(self.resume_file):
                last_checkpoint_s = time.time() - os.path.getmtime(self.resume_file)
            else:
                last_checkpoint_s = (
                    datetime.datetime.now() - self.start_time
                ).total_seconds()
            if last_checkpoint_s > self.check_point_delta_t:
                self.write_current_state()
                self.plot_current_state()
            if getattr(self.sampler, "added_live", False):
                self.sampler._remove_live_points()

        self.sampler.run_nested(**sampler_kwargs)
        self.write_current_state()
        self.plot_current_state()
        return self.sampler.results

    def _remove_checkpoint(self):
        """Remove checkpointed state"""
        if os.path.isfile(self.resume_file):
            os.remove(self.resume_file)

    def read_saved_state(self, continuing=False):
        """
        Read a pickled saved state of the sampler to disk.

        If the live points are present and the run is continuing
        they are removed.
        The random state must be reset, as this isn't saved by the pickle.
        `nqueue` is set to a negative number to trigger the queue to be
        refilled before the first iteration.
        The previous run time is set to self.

        Parameters
        ==========
        continuing: bool
            Whether the run is continuing or terminating, if True, the loaded
            state is mostly written back to disk.
        """
        import dill
        from dynesty import __version__ as dynesty_version

        from ... import __version__ as bilby_version

        versions = dict(bilby=bilby_version, dynesty=dynesty_version)
        if os.path.isfile(self.resume_file):
            logger.info(f"Reading resume file {self.resume_file}")
            with open(self.resume_file, "rb") as file:
                sampler = dill.load(file)

                if not hasattr(sampler, "versions"):
                    logger.warning(
                        f"The resume file {self.resume_file} is corrupted or "
                        "the version of bilby has changed between runs. This "
                        "resume file will be ignored."
                    )
                    return False
                version_warning = (
                    "The {code} version has changed between runs. "
                    "This may cause unpredictable behaviour and/or failure. "
                    "Old version = {old}, new version = {new}."
                )
                for code in versions:
                    if not versions[code] == sampler.versions.get(code, None):
                        logger.warning(
                            version_warning.format(
                                code=code,
                                old=sampler.versions.get(code, "None"),
                                new=versions[code],
                            )
                        )
                del sampler.versions
                self.sampler = sampler
                if getattr(self.sampler, "added_live", False) and continuing:
                    self.sampler._remove_live_points()
                self.sampler.nqueue = -1
                self.start_time = self.sampler.kwargs.pop("start_time")
                self.sampling_time = self.sampler.kwargs.pop("sampling_time")
                self.sampler.pool = self.pool
                if self.pool is not None:
                    self.sampler.M = self.pool.map
                else:
                    self.sampler.M = map
            return True
        else:
            logger.info(f"Resume file {self.resume_file} does not exist.")
            return False

    def write_current_state_and_exit(self, signum=None, frame=None):
        if self.pbar is not None:
            self.pbar = self.pbar.close()
        super(Dynesty, self).write_current_state_and_exit(signum=signum, frame=frame)

    def write_current_state(self):
        """
        Write the current state of the sampler to disk.

        The sampler is pickle dumped using `dill`.
        The sampling time is also stored to get the full CPU time for the run.

        The check of whether the sampler is picklable is to catch an error
        when using pytest. Hopefully, this message won't be triggered during
        normal running.
        """

        import dill
        from dynesty import __version__ as dynesty_version

        from ... import __version__ as bilby_version

        if getattr(self, "sampler", None) is None:
            # Sampler not initialized, not able to write current state
            return

        check_directory_exists_and_if_not_mkdir(self.outdir)
        if hasattr(self, "start_time"):
            self._update_sampling_time()
            self.sampler.kwargs["sampling_time"] = self.sampling_time
            self.sampler.kwargs["start_time"] = self.start_time
        self.sampler.versions = dict(bilby=bilby_version, dynesty=dynesty_version)
        self.sampler.pool = None
        self.sampler.M = map
        if dill.pickles(self.sampler):
            safe_file_dump(self.sampler, self.resume_file, dill)
            logger.info(f"Written checkpoint file {self.resume_file}")
        else:
            logger.warning(
                "Cannot write pickle resume file! "
                "Job will not resume if interrupted."
            )
        self.sampler.pool = self.pool
        if self.sampler.pool is not None:
            self.sampler.M = self.sampler.pool.map

    def dump_samples_to_dat(self):
        """
        Save the current posterior samples to a space-separated plain-text
        file. These are unbiased posterior samples, however, there will not
        be many of them until the analysis is nearly over.
        """
        sampler = self.sampler
        ln_weights = sampler.saved_logwt - sampler.saved_logz[-1]

        weights = np.exp(ln_weights)
        samples = rejection_sample(np.array(sampler.saved_v), weights)
        nsamples = len(samples)

        # If we don't have enough samples, don't dump them
        if nsamples < 100:
            return

        filename = f"{self.outdir}/{self.label}_samples.dat"
        logger.info(f"Writing {nsamples} current samples to {filename}")

        df = DataFrame(samples, columns=self.search_parameter_keys)
        df.to_csv(filename, index=False, header=True, sep=" ")

    def plot_current_state(self):
        """
        Make diagonstic plots of the history and current state of the sampler.

        These plots are a mixture of :code:`dynesty` implemented run and trace
        plots and our custom stats plot. We also make a copy of the trace plot
        using the unit hypercube samples to reflect the internal state of the
        sampler.

        Any errors during plotting should be handled so that sampling can
        continue.
        """
        if self.check_point_plot:
            import dynesty.plotting as dyplot
            import matplotlib.pyplot as plt

            labels = [label.replace("_", " ") for label in self.search_parameter_keys]
            try:
                filename = f"{self.outdir}/{self.label}_checkpoint_trace.png"
                fig = dyplot.traceplot(self.sampler.results, labels=labels)[0]
                fig.tight_layout()
                fig.savefig(filename)
            except (
                RuntimeError,
                np.linalg.linalg.LinAlgError,
                ValueError,
                OverflowError,
            ) as e:
                logger.warning(e)
                logger.warning("Failed to create dynesty state plot at checkpoint")
            except Exception as e:
                logger.warning(
                    f"Unexpected error {e} in dynesty plotting. "
                    "Please report at git.ligo.org/lscsoft/bilby/-/issues"
                )
            finally:
                plt.close("all")
            try:
                filename = f"{self.outdir}/{self.label}_checkpoint_trace_unit.png"
                from copy import deepcopy

                from dynesty.utils import results_substitute

                temp = deepcopy(self.sampler.results)
                temp = results_substitute(temp, dict(samples=temp["samples_u"]))
                fig = dyplot.traceplot(temp, labels=labels)[0]
                fig.tight_layout()
                fig.savefig(filename)
            except (
                RuntimeError,
                np.linalg.linalg.LinAlgError,
                ValueError,
                OverflowError,
            ) as e:
                logger.warning(e)
                logger.warning("Failed to create dynesty unit state plot at checkpoint")
            except Exception as e:
                logger.warning(
                    f"Unexpected error {e} in dynesty plotting. "
                    "Please report at git.ligo.org/lscsoft/bilby/-/issues"
                )
            finally:
                plt.close("all")
            try:
                filename = f"{self.outdir}/{self.label}_checkpoint_run.png"
                fig, _ = dyplot.runplot(
                    self.sampler.results, logplot=False, use_math_text=False
                )
                fig.tight_layout()
                plt.savefig(filename)
            except (
                RuntimeError,
                np.linalg.linalg.LinAlgError,
                ValueError,
                OverflowError,
            ) as e:
                logger.warning(e)
                logger.warning("Failed to create dynesty run plot at checkpoint")
            except Exception as e:
                logger.warning(
                    f"Unexpected error {e} in dynesty plotting. "
                    "Please report at git.ligo.org/lscsoft/bilby/-/issues"
                )
            finally:
                plt.close("all")
            try:
                filename = f"{self.outdir}/{self.label}_checkpoint_stats.png"
                fig, _ = dynesty_stats_plot(self.sampler)
                fig.tight_layout()
                plt.savefig(filename)
            except (RuntimeError, ValueError, OverflowError) as e:
                logger.warning(e)
                logger.warning("Failed to create dynesty stats plot at checkpoint")
            except DynestySetupError:
                logger.debug("Cannot create Dynesty stats plot with dynamic sampler.")
            except Exception as e:
                logger.warning(
                    f"Unexpected error {e} in dynesty plotting. "
                    "Please report at git.ligo.org/lscsoft/bilby/-/issues"
                )
            finally:
                plt.close("all")

    def _run_test(self):
        """Run the sampler very briefly as a sanity test that it works."""
        import pandas as pd

        self._set_sampling_method()
        self._setup_pool()
        self.sampler = self.sampler_init(
            loglikelihood=_log_likelihood_wrapper,
            prior_transform=_prior_transform_wrapper,
            ndim=self.ndim,
            **self.sampler_init_kwargs,
        )
        sampler_kwargs = self.sampler_function_kwargs.copy()
        sampler_kwargs["maxiter"] = 2

        if self.print_method == "tqdm" and self.kwargs["print_progress"]:
            from tqdm.auto import tqdm

            self.pbar = tqdm(file=sys.stdout, initial=self.sampler.it)
        self.sampler.run_nested(**sampler_kwargs)
        self._close_pool()

        if self.pbar is not None:
            self.pbar = self.pbar.close()
            print("")
        N = 100
        self.result.samples = pd.DataFrame(self.priors.sample(N))[
            self.search_parameter_keys
        ].values
        self.result.nested_samples = self.result.samples
        self.result.log_likelihood_evaluations = np.ones(N)
        self.result.log_evidence = 1
        self.result.log_evidence_err = 0.1

        return self.result

    def prior_transform(self, theta):
        """Prior transform method that is passed into the external sampler.
        cube we map this back to [0, 1].

        Parameters
        ==========
        theta: list
            List of sampled values on a unit interval

        Returns
        =======
        list: Properly rescaled sampled values

        """
        return self.priors.rescale(self._search_parameter_keys, theta)


@latex_plot_format
def dynesty_stats_plot(sampler):
    """
    Plot diagnostic statistics from a dynesty run

    The plotted quantities per iteration are:

    - nc: the number of likelihood calls
    - scale: the number of accepted MCMC steps if using :code:`bound="live"`
      or :code:`bound="live-multi"`, otherwise, the scale applied to the MCMC
      steps
    - lifetime: the number of iterations a point stays in the live set

    There is also a histogram of the lifetime compared with the theoretical
    distribution. To avoid edge effects, we discard the first 6 * nlive

    Parameters
    ----------
    sampler: dynesty.sampler.Sampler
        The sampler object containing the run history.

    Returns
    -------
    fig: matplotlib.pyplot.figure.Figure
        Figure handle for the new plot
    axs: matplotlib.pyplot.axes.Axes
        Axes handles for the new plot

    """
    import matplotlib.pyplot as plt
    from scipy.stats import geom, ks_1samp

    fig, axs = plt.subplots(nrows=4, figsize=(8, 8))
    data = sampler.saved_run.D
    for ax, name in zip(axs, ["nc", "scale"]):
        ax.plot(data[name], color="blue")
        ax.set_ylabel(name.title())
    lifetimes = np.arange(len(data["it"])) - data["it"]
    axs[-2].set_ylabel("Lifetime")
    if not hasattr(sampler, "nlive"):
        raise DynestySetupError("Cannot make stats plot for dynamic sampler.")
    nlive = sampler.nlive
    burn = int(geom(p=1 / nlive).isf(1 / 2 / nlive))
    if len(data["it"]) > burn + sampler.nlive:
        axs[-2].plot(np.arange(0, burn), lifetimes[:burn], color="grey")
        axs[-2].plot(
            np.arange(burn, len(lifetimes) - nlive),
            lifetimes[burn:-nlive],
            color="blue",
        )
        axs[-2].plot(
            np.arange(len(lifetimes) - nlive, len(lifetimes)),
            lifetimes[-nlive:],
            color="red",
        )
        lifetimes = lifetimes[burn:-nlive]
        ks_result = ks_1samp(lifetimes, geom(p=1 / nlive).cdf)
        axs[-1].hist(
            lifetimes,
            bins=np.linspace(0, 6 * nlive, 60),
            histtype="step",
            density=True,
            color="blue",
            label=f"p value = {ks_result.pvalue:.3f}",
        )
        axs[-1].plot(
            np.arange(1, 6 * nlive),
            geom(p=1 / nlive).pmf(np.arange(1, 6 * nlive)),
            color="red",
        )
        axs[-1].set_xlim(0, 6 * nlive)
        axs[-1].legend()
        axs[-1].set_yscale("log")
    else:
        axs[-2].plot(
            np.arange(0, len(lifetimes) - nlive), lifetimes[:-nlive], color="grey"
        )
        axs[-2].plot(
            np.arange(len(lifetimes) - nlive, len(lifetimes)),
            lifetimes[-nlive:],
            color="red",
        )
    axs[-2].set_yscale("log")
    axs[-2].set_xlabel("Iteration")
    axs[-1].set_xlabel("Lifetime")
    return fig, axs


class DynestySetupError(Exception):
    pass



#!/usr/bin/env python
"""
Tutorial to demonstrate running parameter estimation on a binary neutron star
system taking into account tidal deformabilities with a physically motivated
model for the tidal deformabilities.

WARNING: The code is extremely slow.
"""


import bilby
import numpy as np
from bilby.gw.eos import EOSFamily, TabularEOS

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "bns_eos_example"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(88170235)

# We are going to inject a binary neutron star waveform.  We first establish a
# dictionary of parameters that includes all of the different waveform
# parameters, including masses of the two black holes (mass_1, mass_2),
# aligned spins of both black holes (chi_1, chi_2), etc.

# We're also going to 'inject' the MPA1 equation of state.
# This is done by injecting masses for the two neutron-stars,
# assuming a specific equation of state, and calculating
# corresponding tidal deformability parameters from the EoS and
# masses.
mpa1_eos = TabularEOS("MPA1")
mpa1_fam = EOSFamily(mpa1_eos)

mass_1 = 1.5
mass_2 = 1.3
lambda_1 = mpa1_fam.lambda_from_mass(mass_1)
lambda_2 = mpa1_fam.lambda_from_mass(mass_2)


injection_parameters = dict(
    mass_1=mass_1,
    mass_2=mass_2,
    chi_1=0.02,
    chi_2=0.02,
    luminosity_distance=50.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
    lambda_1=lambda_1,
    lambda_2=lambda_2,
)

# Set the duration and sampling frequency of the data segment that we're going
# to inject the signal into. For the
# TaylorF2 waveform, we cut the signal close to the isco frequency
duration = 32
sampling_frequency = 2048
start_time = injection_parameters["geocent_time"] + 2 - duration

# Fixed arguments passed into the source model. The analysis starts at 40 Hz.
# Note that the EoS sampling is agnostic to waveform model as long as the approximant
# can include tides.
waveform_arguments = dict(
    waveform_approximant="IMRPhenomPv2_NRTidal",
    reference_frequency=50.0,
    minimum_frequency=40.0,
)

# Create the waveform_generator using a LAL Binary Neutron Star source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_neutron_star,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters,
    waveform_arguments=waveform_arguments,
)

# Set up interferometers.  In this case we'll use three interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1), and Virgo (V1)).
# These default to their design sensitivity and start at 40 Hz.
interferometers = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
for interferometer in interferometers:
    interferometer.minimum_frequency = 40
interferometers.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency, duration=duration, start_time=start_time
)
interferometers.inject_signal(
    parameters=injection_parameters, waveform_generator=waveform_generator
)

# We're going to sample in chirp_mass, symmetric_mass_ratio, and
# specific EoS model parameters. We're using a 4-parameter
# spectrally-decomposed EoS parameterization from Lindblom (2010).
# BNS have aligned spins by default, if you want to allow precessing spins
# pass aligned_spin=False to the BNSPriorDict
priors = bilby.gw.prior.BNSPriorDict()
for key in [
    "psi",
    "geocent_time",
    "ra",
    "dec",
    "chi_1",
    "chi_2",
    "theta_jn",
    "luminosity_distance",
    "phase",
]:
    priors[key] = injection_parameters[key]
for key in ["mass_1", "mass_2", "lambda_1", "lambda_2", "mass_ratio"]:
    del priors[key]
priors["chirp_mass"] = bilby.core.prior.Gaussian(
    1.215, 0.1, name="chirp_mass", unit="$M_{\\odot}$"
)
priors["symmetric_mass_ratio"] = bilby.core.prior.Uniform(
    0.1, 0.25, name="symmetric_mass_ratio"
)
priors["eos_spectral_gamma_0"] = bilby.core.prior.Uniform(
    0.2, 2.0, name="gamma0", latex_label="$\\gamma_0"
)
priors["eos_spectral_gamma_1"] = bilby.core.prior.Uniform(
    -1.6, 1.7, name="gamma1", latex_label="$\\gamma_1"
)
priors["eos_spectral_gamma_2"] = bilby.core.prior.Uniform(
    -0.6, 0.6, name="gamma2", latex_label="$\\gamma_2"
)
priors["eos_spectral_gamma_3"] = bilby.core.prior.Uniform(
    -0.02, 0.02, name="gamma3", latex_label="$\\gamma_3"
)

# The eos_check prior imposes several hard physical constraints on samples like
# enforcing causality and monotinicity of the EoSs. In almost ever conceivable
# sampling scenario, this should be enabled.
priors["eos_check"] = bilby.gw.prior.EOSCheck()

# Initialise the likelihood by passing in the interferometer data (IFOs)
# and the waveform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=interferometers,
    waveform_generator=waveform_generator,
)

# Run sampler.  In this case we're going to use the `dynesty` sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    npoints=1000,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
    conversion_function=bilby.gw.conversion.generate_all_bns_parameters,
    resume=True,
)

result.plot_corner()



#!/usr/bin/env python
"""
An example of how to use bilby to perform parameter estimation for
fitting a linear function to data with background Gaussian noise.
This will compare the output of using a stochastic sampling method
to evaluating the posterior on a grid.
"""
import bilby
import matplotlib.pyplot as plt
import numpy as np

# A few simple setup steps
label = "linear_regression_grid"
outdir = "outdir"
bilby.utils.check_directory_exists_and_if_not_mkdir(outdir)


# First, we define our "signal model", in this case a simple linear function
def model(time, m, c):
    return time * m + c


# Now we define the injection parameters which we make simulated data with
injection_parameters = dict()
injection_parameters["c"] = 0.2
injection_parameters["m"] = 0.5

# For this example, we'll use standard Gaussian noise

# These lines of code generate the fake data. Note the ** just unpacks the
# contents of the injection_parameters when calling the model function.
sampling_frequency = 10
time_duration = 10
time = np.arange(0, time_duration, 1 / sampling_frequency)
N = len(time)
sigma = 3.0
data = model(time, **injection_parameters) + np.random.normal(0, sigma, N)

# We quickly plot the data to check it looks sensible
fig, ax = plt.subplots()
ax.plot(time, data, "o", label="data")
ax.plot(time, model(time, **injection_parameters), "--r", label="signal")
ax.set_xlabel("time")
ax.set_ylabel("y")
ax.legend()
fig.savefig("{}/{}_data.png".format(outdir, label))

# Now lets instantiate a version of our GaussianLikelihood, giving it
# the time, data and signal model
likelihood = bilby.likelihood.GaussianLikelihood(time, data, model, sigma)

# From hereon, the syntax is exactly equivalent to other bilby examples
# We make a prior
priors = bilby.core.prior.PriorDict()
priors["m"] = bilby.core.prior.Uniform(0, 5, "m")
priors["c"] = bilby.core.prior.Uniform(-2, 2, "c")

# And run sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    nlive=500,
    sample="unif",
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)
fig = result.plot_corner(parameters=injection_parameters, save=False)

grid = bilby.core.grid.Grid(likelihood, priors, grid_size={"m": 200, "c": 100})

# overplot the grid estimates
grid_evidence = grid.log_evidence
axes = fig.get_axes()
axes[0].plot(
    grid.sample_points["c"],
    np.exp(grid.marginalize_ln_posterior(not_parameters="c") - grid_evidence),
    "k--",
)
axes[3].plot(
    grid.sample_points["m"],
    np.exp(grid.marginalize_ln_posterior(not_parameters="m") - grid_evidence),
    "k--",
)
axes[2].contour(
    grid.mesh_grid[1],
    grid.mesh_grid[0],
    np.exp(grid.ln_posterior - np.max(grid.ln_posterior)),
)

fig.savefig("{}/{}_corner.png".format(outdir, label), dpi=300)

# compare evidences
print("Dynesty log(evidence): {}".format(result.log_evidence))
print("Grid log(evidence): {}".format(grid_evidence))



#!/usr/bin/env python
"""
This example demonstrates how to simulate some data, add an injected signal
and plot the data.
"""
import numpy as np
from bilby.gw.detector import get_empty_interferometer
from bilby.gw.source import lal_binary_black_hole
from bilby.gw.waveform_generator import WaveformGenerator

np.random.seed(1)

duration = 4
sampling_frequency = 2048.0

outdir = "outdir"
label = "example"

injection_parameters = dict(
    mass_1=36.0,
    mass_2=29.0,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=1000.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)

waveform_arguments = dict(
    waveform_approximant="IMRPhenomTPHM", reference_frequency=50.0
)

waveform_generator = WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=lal_binary_black_hole,
    parameters=injection_parameters,
    waveform_arguments=waveform_arguments,
)
hf_signal = waveform_generator.frequency_domain_strain(injection_parameters)

ifo = get_empty_interferometer("H1")
ifo.set_strain_data_from_power_spectral_density(
    duration=duration, sampling_frequency=sampling_frequency
)
ifo.inject_signal(injection_polarizations=hf_signal, parameters=injection_parameters)

t0 = injection_parameters["geocent_time"]
ifo.plot_time_domain_data(
    outdir=outdir,
    label=label,
    notches=[50],
    bandpass_frequencies=(50, 200),
    start_end=(-0.5, 0.5),
    t0=t0,
)



import unittest
from copy import deepcopy

import bilby
import numpy as np
from parameterized import parameterized


class TestRelativeBinningLikelihood(unittest.TestCase):
    def setUp(self):
        duration = 16
        fmin = 20
        sampling_frequency = 8192
        chirp_mass = 13
        mass_ratio = 0.5
        mass_1, mass_2 = bilby.gw.conversion.chirp_mass_and_mass_ratio_to_component_masses(
            chirp_mass=chirp_mass, mass_ratio=mass_ratio
        )
        self.test_parameters = dict(
            chirp_mass=13,
            mass_ratio=0.5,
            a_1=0.3,
            a_2=0.4,
            tilt_1=1.0,
            tilt_2=0.2,
            phi_12=1.0,
            phi_jl=2.0,
            luminosity_distance=2000.0,
            theta_jn=0.4,
            psi=0.659,
            phase=1.3,
            geocent_time=1187008882,
            ra=1.3,
            dec=-1.2,
        )
        self.fiducial_parameters = self.test_parameters.copy()
        del self.fiducial_parameters["chirp_mass"], self.fiducial_parameters["mass_ratio"]
        self.fiducial_parameters["mass_1"] = mass_1
        self.fiducial_parameters["mass_2"] = mass_2

        ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration,
            start_time=self.test_parameters['geocent_time'] - duration + 2.
        )
        for ifo in ifos:
            ifo.minimum_frequency = fmin

        spline_calibration_nodes = 10
        self.calibration_parameters = {}
        for ifo in ifos:
            ifo.calibration_model = bilby.gw.calibration.CubicSpline(
                prefix=f"recalib_{ifo.name}_",
                minimum_frequency=ifo.minimum_frequency,
                maximum_frequency=ifo.maximum_frequency,
                n_points=spline_calibration_nodes
            )
            for i in range(spline_calibration_nodes):
                self.test_parameters[f"recalib_{ifo.name}_amplitude_{i}"] = 0
                self.test_parameters[f"recalib_{ifo.name}_phase_{i}"] = 0
                # Calibration errors of 5% in amplitude and 5 degrees in phase
                self.calibration_parameters[f"recalib_{ifo.name}_amplitude_{i}"] = \
                    np.random.normal(loc=0, scale=0.05)
                self.calibration_parameters[f"recalib_{ifo.name}_phase_{i}"] = \
                    np.random.normal(loc=0, scale=5 * np.pi / 180)

        priors = bilby.gw.prior.BBHPriorDict()
        priors.pop("mass_1")
        priors.pop("mass_2")
        priors["chirp_mass"] = bilby.core.prior.Uniform(12, 14)
        priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        priors["geocent_time"] = self.test_parameters["geocent_time"]
        priors["fiducial"] = 0
        self.priors = priors

        approximant = "IMRPhenomXP"
        non_bin_wfg = bilby.gw.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            waveform_arguments=dict(
                reference_frequency=fmin, minimum_frequency=fmin, approximant=approximant)
        )
        bin_wfg = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole_relative_binning,
            waveform_arguments=dict(
                reference_frequency=fmin, approximant=approximant, minimum_frequency=fmin)
        )
        ifos.inject_signal(
            parameters=self.test_parameters,
            waveform_generator=non_bin_wfg,
            raise_error=False,
        )
        self.ifos = ifos

        self.non_bin = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos, waveform_generator=deepcopy(non_bin_wfg),
            priors=priors.copy()
        )
        self.binned = bilby.gw.likelihood.RelativeBinningGravitationalWaveTransient(
            interferometers=ifos, waveform_generator=deepcopy(bin_wfg),
            fiducial_parameters=self.fiducial_parameters,
            priors=priors.copy(),
            epsilon=0.05,
        )
        self.non_bin.parameters.update(self.test_parameters)
        self.reference_ln_l = self.non_bin.log_likelihood_ratio()
        self.bin_wfg = bin_wfg
        self.priors = priors

    def tearDown(self):
        del (
            self.non_bin,
            self.binned,
        )

    def test_matches_non_binned_many(self):
        for _ in range(100):
            parameters = self.priors.sample()
            self.non_bin.parameters.update(parameters)
            self.binned.parameters.update(parameters)
            regular_ln_l = self.non_bin.log_likelihood_ratio()
            binned_ln_l = self.binned.log_likelihood_ratio()
            self.assertLess(
                abs(regular_ln_l - binned_ln_l)
                / abs(self.reference_ln_l - regular_ln_l),
                0.1
            )

    @parameterized.expand([(False, ), (True, )])
    def test_matches_non_binned(self, add_cal_errors):
        self.non_bin.parameters.update(self.test_parameters)
        self.binned.parameters.update(self.test_parameters)
        if add_cal_errors:
            self.non_bin.parameters.update(self.calibration_parameters)
            self.binned.parameters.update(self.calibration_parameters)
        regular_ln_l = self.non_bin.log_likelihood_ratio()
        binned_ln_l = self.binned.log_likelihood_ratio()
        self.assertLess(abs(regular_ln_l - binned_ln_l), 1e-3)

    def test_optimization_gives_good_match(self):
        fiducial_parameters = self.test_parameters.copy()
        fiducial_parameters["chirp_mass"] *= 0.99
        priors = self.priors.copy()
        for key in [
            "ra", "dec", "geocent_time", "phase", "psi", "theta_jn", "luminosity_distance",
            "a_1", "a_2", "tilt_1", "tilt_2", "phi_12", "phi_jl",
        ]:
            priors[key] = self.test_parameters[key]
        binned = bilby.gw.likelihood.RelativeBinningGravitationalWaveTransient(
            interferometers=self.ifos, waveform_generator=deepcopy(self.bin_wfg),
            priors=priors,
            fiducial_parameters=fiducial_parameters,
            epsilon=0.05,
            update_fiducial_parameters=True,
        )
        self.non_bin.parameters.update(self.test_parameters)
        binned.parameters.update(self.test_parameters)
        regular_ln_l = self.non_bin.log_likelihood_ratio()
        binned_ln_l = binned.log_likelihood_ratio()
        self.assertLess(abs(regular_ln_l - binned_ln_l), 1e-3)

    def test_very_small_epsilon_returns_good_value(self):
        """
        If the frequency bins cover less than one bin, the likeilhood is nan,
        test that we avoid this.
        """
        binned = bilby.gw.likelihood.RelativeBinningGravitationalWaveTransient(
            interferometers=self.ifos, waveform_generator=deepcopy(self.bin_wfg),
            fiducial_parameters=self.fiducial_parameters,
            priors=self.priors.copy(),
            epsilon=0.001,
        )
        binned.parameters.update(self.test_parameters)
        self.assertFalse(np.isnan(binned.log_likelihood_ratio()))


if __name__ == "__main__":
    unittest.main()



from datetime import date

with open("dockerfile-template", "r") as ff:
    template = ff.read()

python_versions = [(3, 8), (3, 9), (3, 10)]
today = date.today().strftime("%Y%m%d")

samplers = [
    "dynesty", "emcee", "nestle", "ptemcee", "pymultinest", "ultranest",
    "cpnest", "kombine", "dnest4", "zeus-mcmc",
    "pytorch", "'pymc>=4'", "nessai", "ptmcmcsampler",
]

for python_major_version, python_minor_version in python_versions:
    key = f"python{python_major_version}{python_minor_version}"
    with open(
        f"v3-dockerfile-test-suite-{key}",
        "w"
    ) as ff:
        ff.write(
            "# This dockerfile is written automatically and should not be "
            "modified by hand.\n\n"
        )
        ff.write(template.format(
            date=today,
            python_major_version=python_major_version,
            python_minor_version=python_minor_version,
            conda_samplers=" ".join(samplers)
        ))



import os
import unittest

import numpy as np
from bilby.core.prior import PriorDict
from bilby.gw import calibration, detector, prior


class TestBaseClass(unittest.TestCase):
    def setUp(self):
        self.model = calibration.Recalibrate()

    def tearDown(self):
        del self.model

    def test_repr(self):
        expected = "Recalibrate(prefix={})".format("'recalib_'")
        actual = repr(self.model)
        self.assertEqual(expected, actual)

    def test_calibration_factor(self):
        frequency_array = np.linspace(20, 1024, 1000)
        cal_factor = self.model.get_calibration_factor(frequency_array)
        assert np.alltrue(cal_factor.real == np.ones_like(frequency_array))


class TestCubicSpline(unittest.TestCase):
    def setUp(self):
        self.prefix = "recalib_"
        self.minimum_frequency = 20
        self.maximum_frequency = 1024
        self.n_points = 5
        self.model = calibration.CubicSpline(
            prefix=self.prefix,
            minimum_frequency=self.minimum_frequency,
            maximum_frequency=self.maximum_frequency,
            n_points=self.n_points,
        )
        self.parameters = {
            "recalib_{}_{}".format(param, ii): 0.0
            for ii in range(5)
            for param in ["amplitude", "phase"]
        }

    def tearDown(self):
        del self.prefix
        del self.minimum_frequency
        del self.maximum_frequency
        del self.n_points
        del self.model
        del self.parameters

    def test_calibration_factor(self):
        frequency_array = np.linspace(20, 1024, 1000)
        cal_factor = self.model.get_calibration_factor(
            frequency_array, **self.parameters
        )
        assert np.alltrue(cal_factor.real == np.ones_like(frequency_array))

    def test_repr(self):
        expected = "CubicSpline(prefix='{}', minimum_frequency={}, maximum_frequency={}, n_points={})".format(
            self.prefix, self.minimum_frequency, self.maximum_frequency, self.n_points
        )
        actual = repr(self.model)
        self.assertEqual(expected, actual)


class TestCubicSplineRequiresFourNodes(unittest.TestCase):
    def test_cannot_instantiate_with_too_few_nodes(self):
        for ii in range(6):
            if ii < 4:
                with self.assertRaises(ValueError):
                    calibration.CubicSpline("test", 1, 10, ii)
            else:
                calibration.CubicSpline("test", 1, 10, ii)


class TestReadWriteCalibrationDraws(unittest.TestCase):
    def setUp(self):
        self.filename = "calibration_draws.h5"
        self.number_of_draws = 100
        self.ifo = detector.get_empty_interferometer("H1")
        self.ifo.set_strain_data_from_power_spectral_density(
            duration=1, sampling_frequency=512
        )
        self.ifo.calibration_model = calibration.CubicSpline(
            prefix="recalib_H1_",
            n_points=10,
            minimum_frequency=self.ifo.minimum_frequency,
            maximum_frequency=self.ifo.maximum_frequency,
        )
        self.priors = prior.CalibrationPriorDict.constant_uncertainty_spline(
            amplitude_sigma=0.1,
            phase_sigma=0.1,
            minimum_frequency=self.ifo.minimum_frequency,
            maximum_frequency=self.ifo.maximum_frequency,
            n_nodes=10,
            label="H1",
        )

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_generate_draws(self):
        draws, parameters = calibration._generate_calibration_draws(
            self.ifo, self.priors, self.number_of_draws
        )
        self.assertEqual(draws.shape, (self.number_of_draws, sum(self.ifo.frequency_mask)))
        self.assertListEqual(list(self.priors.keys()), list(parameters.keys()))

    def test_read_write_matches(self):
        draws, parameters = calibration._generate_calibration_draws(
            self.ifo, self.priors, self.number_of_draws
        )
        frequencies = self.ifo.frequency_array[self.ifo.frequency_mask]
        calibration.write_calibration_file(
            filename=self.filename,
            frequency_array=frequencies,
            calibration_draws=draws,
            calibration_parameter_draws=parameters,
        )
        self.assertTrue(os.path.exists(self.filename))
        loaded_draws, loaded_parameters = calibration.read_calibration_file(
            filename=self.filename,
            frequency_array=frequencies,
            number_of_response_curves=self.number_of_draws,
        )
        self.assertLess(np.max(np.abs(loaded_draws - draws)), 1e-13)
        self.assertTrue(parameters.equals(loaded_parameters))

    def test_build_calibration_lookup(self):
        ifos = detector.InterferometerList(["H1", "L1", "H1"])
        ifos.set_strain_data_from_power_spectral_densities(
            duration=4, sampling_frequency=1024
        )
        priors = PriorDict()
        for ifo in ifos:
            ifo.minimum_frequency = 20
            ifo.maximum_frequency = 1024
            priors.update(prior.CalibrationPriorDict.constant_uncertainty_spline(
                amplitude_sigma=0.1,
                phase_sigma=0.1,
                minimum_frequency=ifo.minimum_frequency,
                maximum_frequency=ifo.maximum_frequency,
                n_nodes=10,
                label=ifo.name,
            ))
            ifo.calibration_model = calibration.CubicSpline(
                prefix=f"recalib_{ifo.name}_",
                n_points=10,
                minimum_frequency=ifo.minimum_frequency,
                maximum_frequency=ifo.maximum_frequency,
            )
        draws, parameters = calibration.build_calibration_lookup(
            interferometers=ifos,
            priors=priors,
            number_of_response_curves=10,
        )
        parameters
        for ifo in ifos[:2]:
            self.assertIn(ifo.name, draws)
            self.assertIn(ifo.name, parameters)
            filename = f"{ifo.name}_calibration_file.h5"
            self.assertTrue(os.path.exists(filename))
            os.remove(filename)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
An example of how to use bilby to perform parameter estimation for
non-gravitational wave data consisting of a Gaussian with a mean and variance
"""
import bilby
import numpy as np

# A few simple setup steps
label = "gaussian_example"
outdir = "outdir"

# Here is minimum requirement for a Likelihood class to run with bilby. In this
# case, we setup a GaussianLikelihood, which needs to have a log_likelihood
# method. Note, in this case we will NOT make use of the `bilby`
# waveform_generator to make the signal.

# Making simulated data: in this case, we consider just a Gaussian

data = np.random.normal(3, 4, 100)


class SimpleGaussianLikelihood(bilby.Likelihood):
    def __init__(self, data):
        """
        A very simple Gaussian likelihood

        Parameters
        ----------
        data: array_like
            The data to analyse
        """
        super().__init__(parameters={"mu": None, "sigma": None})
        self.data = data
        self.N = len(data)

    def log_likelihood(self):
        mu = self.parameters["mu"]
        sigma = self.parameters["sigma"]
        res = self.data - mu
        return -0.5 * (
            np.sum((res / sigma) ** 2) + self.N * np.log(2 * np.pi * sigma**2)
        )


likelihood = SimpleGaussianLikelihood(data)
priors = dict(
    mu=bilby.core.prior.Uniform(0, 5, "mu"),
    sigma=bilby.core.prior.Uniform(0, 10, "sigma"),
)

# And run sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    nlive=1000,
    outdir=outdir,
    label=label,
)
result.plot_corner()



#!/usr/bin/env python
"""
Tutorial to demonstrates reading and plotting the GraceDB data for GW170608,
the lightest GWTC-1 BBH event.

This tutorial must be run on a LIGO cluster.
"""

import bilby

label = "GW170608"
outdir = label + "_out"
# List of detectors involved in the event. GW170608 is pre-August 2017 so only
# H1 and L1 were involved
detectors = ["H1", "L1"]
# Names of the channels and query types to use, and the event GraceDB ID - ref.
# https://ldas-jobs.ligo.caltech.edu/~eve.chase/monitor/online_pe/C02_clean/
# C02_HL_Pv2/lalinferencenest/IMRPhenomPv2pseudoFourPN/config.ini
channel_names = ["H1:DCH-CLEAN_STRAIN_C02", "L1:DCH-CLEAN_STRAIN_C02"]
query_types = ["H1_CLEANED_HOFT_C02", "L1_CLEANED_HOFT_C02"]
gracedb = "G288686"
# Calibration number to inform bilby's guess of query type if those provided are
# not recognised
calibration = 2
# Duration of data around the event to use
duration = 16
# Duration of PSD data
psd_duration = 1024
# Time of event, stored in bilby
trigger_time = bilby.gw.utils.get_event_time(label)
# Minimum frequency and reference frequency
minimum_frequency = 10  # Hz

# Get frame caches
candidate, frame_caches = bilby.gw.utils.get_gracedb(
    gracedb, outdir, duration, calibration, detectors, query_types
)

# Set up interferometer objects from the cache files
interferometers = bilby.gw.detector.InterferometerList([])

for cache_file, channel_name in zip(frame_caches, channel_names):
    interferometer = bilby.gw.detector.load_data_from_cache_file(
        cache_file, trigger_time, duration, psd_duration, channel_name
    )
    interferometer.minimum_frequency = minimum_frequency
    interferometers.append(interferometer)

# Plot the strain data
interferometers.plot_data(outdir=outdir, label=label)



#!/usr/bin/env python
"""

Tutorial to demonstrate running parameter estimation/model selection on an NR
supernova injected signal.  Signal model is made by applying PCA to a set of
supernova waveforms. The first few PCs are then linearly combined with a scale
factor. (See https://arxiv.org/pdf/1202.3256.pdf)

For this example we use `PyMultiNest`, this can be installed using

conda install -c conda-forge pymultinest
"""
import bilby
import numpy as np

# Set the duration and sampling frequency of the data segment that we're going
# to inject the signal into.
# These are fixed by the resolution in the injection file that we are using.
duration = 3
sampling_frequency = 4096

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "supernova"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(170801)

# We are going to inject a supernova waveform.  We first establish a dictionary
# of parameters that includes all of the different waveform parameters. It will
# read in a signal to inject from a txt file.

injection_parameters = dict(
    luminosity_distance=7.0,
    ra=4.6499,
    dec=-0.5063,
    geocent_time=1126259642.413,
    psi=2.659,
)

# Create the waveform_generator using a supernova source function
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.supernova,
    parameters=injection_parameters,
    parameter_conversion=lambda parameters: (parameters, list()),
    waveform_arguments=dict(file_path="MuellerL15_example_inj.txt"),
)

# Set up interferometers.  In this case we'll use three interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1), and Virgo (V1)).  These default to
# their design sensitivity
ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 1.5,
)
ifos.inject_signal(
    waveform_generator=waveform_generator,
    parameters=injection_parameters,
    raise_error=False,
)

# read in from a file the PCs used to create the signal model.
realPCs = np.genfromtxt("SupernovaRealPCs.txt")
imagPCs = np.genfromtxt("SupernovaImagPCs.txt")

# Now we make another waveform_generator because the signal model is
# not the same as the injection in this case.
simulation_parameters = dict(realPCs=realPCs, imagPCs=imagPCs)

search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.supernova_pca_model,
    waveform_arguments=simulation_parameters,
)

# Set up prior
priors = bilby.core.prior.PriorDict()
for key in ["psi", "geocent_time"]:
    priors[key] = injection_parameters[key]
priors["luminosity_distance"] = bilby.core.prior.Uniform(
    2, 20, "luminosity_distance", unit="$kpc$"
)
priors["pc_coeff1"] = bilby.core.prior.Uniform(-1, 1, "pc_coeff1")
priors["pc_coeff2"] = bilby.core.prior.Uniform(-1, 1, "pc_coeff2")
priors["pc_coeff3"] = bilby.core.prior.Uniform(-1, 1, "pc_coeff3")
priors["pc_coeff4"] = bilby.core.prior.Uniform(-1, 1, "pc_coeff4")
priors["pc_coeff5"] = bilby.core.prior.Uniform(-1, 1, "pc_coeff5")
priors["ra"] = bilby.core.prior.Uniform(
    minimum=0, maximum=2 * np.pi, name="ra", boundary="periodic"
)
priors["dec"] = bilby.core.prior.Sine(name="dec")
priors["geocent_time"] = bilby.core.prior.Uniform(
    injection_parameters["geocent_time"] - 1,
    injection_parameters["geocent_time"] + 1,
    "geocent_time",
    unit="$s$",
)

# Initialise the likelihood by passing in the interferometer data (IFOs) and
# the waveoform generator
likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=search_waveform_generator
)

# Run sampler.
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="pymultinest",
    npoints=500,
    outdir=outdir,
    label=label,
)

# make some plots of the outputs
result.plot_corner()



#!/usr/bin/env python
"""
An example of how to use bilby to perform parameter estimation for hyper params
"""
import matplotlib.pyplot as plt
import numpy as np
from bilby.core.likelihood import GaussianLikelihood
from bilby.core.prior import Uniform
from bilby.core.result import make_pp_plot
from bilby.core.sampler import run_sampler
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from bilby.hyper.likelihood import HyperparameterLikelihood

outdir = "outdir"
check_directory_exists_and_if_not_mkdir(outdir)


# Define a model to fit to each data set
def model(x, c0, c1):
    return c0 + c1 * x


N = 10
x = np.linspace(0, 10, N)
sigma = 1
Nevents = 4
labels = ["a", "b", "c", "d"]

true_mu_c0 = 5
true_sigma_c0 = 1

fig1, ax1 = plt.subplots()
fig2, ax2 = plt.subplots()

# Make the sample sets
results = list()
for i in range(Nevents):
    c0 = np.random.normal(true_mu_c0, true_sigma_c0)
    c1 = np.random.uniform(-1, 1)
    injection_parameters = dict(c0=c0, c1=c1)

    data = model(x, **injection_parameters) + np.random.normal(0, sigma, N)
    line = ax1.plot(x, data, "-x", label=labels[i])

    likelihood = GaussianLikelihood(x, data, model, sigma)
    priors = dict(c0=Uniform(-10, 10, "c0"), c1=Uniform(-10, 10, "c1"))

    result = run_sampler(
        likelihood=likelihood,
        priors=priors,
        sampler="nestle",
        nlive=1000,
        outdir=outdir,
        verbose=False,
        label="individual_{}".format(i),
        save=False,
        injection_parameters=injection_parameters,
    )
    ax2.hist(
        result.posterior.c0,
        color=line[0].get_color(),
        density=True,
        alpha=0.5,
        label=labels[i],
    )
    results.append(result)

ax1.set_xlabel("x")
ax1.set_ylabel("y(x)")
ax1.legend()
fig1.savefig("outdir/hyper_parameter_data.png")
ax2.set_xlabel("c0")
ax2.set_ylabel("density")
ax2.legend()
fig2.savefig("outdir/hyper_parameter_combined_posteriors.png")


def hyper_prior(dataset, mu, sigma):
    return (
        np.exp(-((dataset["c0"] - mu) ** 2) / (2 * sigma**2))
        / (2 * np.pi * sigma**2) ** 0.5
    )


samples = [result.posterior for result in results]
for sample in samples:
    sample["prior"] = 1 / 20
evidences = [result.log_evidence for result in results]
hp_likelihood = HyperparameterLikelihood(
    posteriors=samples,
    hyper_prior=hyper_prior,
    log_evidences=evidences,
    max_samples=500,
)

hp_priors = dict(
    mu=Uniform(-10, 10, "mu", r"$\mu_{c0}$"),
    sigma=Uniform(0, 10, "sigma", r"$\sigma_{c0}$"),
)

# And run sampler
result = run_sampler(
    likelihood=hp_likelihood,
    priors=hp_priors,
    sampler="dynesty",
    nlive=1000,
    use_ratio=False,
    outdir=outdir,
    label="hyper_parameter",
    verbose=True,
    clean=True,
)
result.plot_corner(truth=dict(mu=true_mu_c0, sigma=true_sigma_c0))
make_pp_plot(results, filename=outdir + "/hyper_parameter_pp.png")



#!/usr/bin/env python
"""
Example of how to use the Reduced Order Quadrature method (see Smith et al.,
(2016) Phys. Rev. D 94, 044031) for a Binary Black hole simulated signal in
Gaussian noise.

This requires files specifying the appropriate basis weights.
These aren't shipped with Bilby, but are available on LDG clusters and
from the public repository https://git.ligo.org/lscsoft/ROQ_data.

We also reweight the result using the regular waveform model to check how
correct the ROQ result is.
"""

import bilby
import numpy as np

outdir = "outdir"
label = "roq"

# The ROQ bases can be given an overall scaling.
# Note how this is applied to durations, frequencies and masses.
scale_factor = 1.6

# Load in the pieces for the linear part of the ROQ. Note you will need to
# adjust the filenames here to the correct paths on your machine
basis_matrix_linear = np.load("B_linear.npy").T
freq_nodes_linear = np.load("fnodes_linear.npy") * scale_factor

# Load in the pieces for the quadratic part of the ROQ
basis_matrix_quadratic = np.load("B_quadratic.npy").T
freq_nodes_quadratic = np.load("fnodes_quadratic.npy") * scale_factor

# Load the parameters describing the valid parameters for the basis.
params = np.genfromtxt("params.dat", names=True)

# Get scaled ROQ quantities
minimum_chirp_mass = params["chirpmassmin"] / scale_factor
maximum_chirp_mass = params["chirpmassmax"] / scale_factor
minimum_component_mass = params["compmin"] / scale_factor

np.random.seed(170808)

duration = 4 / scale_factor
sampling_frequency = 2048 * scale_factor

injection_parameters = dict(
    mass_1=36.0,
    mass_2=29.0,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.0,
    tilt_2=0.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=1000.0,
    theta_jn=0.4,
    psi=0.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)

waveform_arguments = dict(
    waveform_approximant="IMRPhenomPv2", reference_frequency=20.0 * scale_factor
)

waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    waveform_arguments=waveform_arguments,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
)

ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 2 / scale_factor,
)
ifos.inject_signal(
    waveform_generator=waveform_generator, parameters=injection_parameters
)
for ifo in ifos:
    ifo.minimum_frequency = 20 * scale_factor

# make ROQ waveform generator
search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
    waveform_arguments=dict(
        frequency_nodes_linear=freq_nodes_linear,
        frequency_nodes_quadratic=freq_nodes_quadratic,
        reference_frequency=20.0 * scale_factor,
        waveform_approximant="IMRPhenomPv2",
    ),
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
)

# Here we add constraints on chirp mass and mass ratio to the prior, these are
# determined by the domain of validity of the ROQ basis.
priors = bilby.gw.prior.BBHPriorDict()
for key in [
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "theta_jn",
    "phase",
    "psi",
    "ra",
    "dec",
    "phi_12",
    "phi_jl",
    "luminosity_distance",
]:
    priors[key] = injection_parameters[key]
for key in ["mass_1", "mass_2"]:
    priors[key].minimum = max(priors[key].minimum, minimum_component_mass)
priors["chirp_mass"] = bilby.core.prior.Uniform(
    name="chirp_mass",
    minimum=float(minimum_chirp_mass),
    maximum=float(maximum_chirp_mass),
)
# The roq parameters typically store the mass ratio bounds as m1/m2 not m2/m1 as in the
# Bilby convention.
priors["mass_ratio"] = bilby.core.prior.Uniform(
    1 / params["qmax"], 1, name="mass_ratio"
)
priors["geocent_time"] = bilby.core.prior.Uniform(
    injection_parameters["geocent_time"] - 0.1,
    injection_parameters["geocent_time"] + 0.1,
    latex_label="$t_c$",
    unit="s",
)

likelihood = bilby.gw.likelihood.ROQGravitationalWaveTransient(
    interferometers=ifos,
    waveform_generator=search_waveform_generator,
    linear_matrix=basis_matrix_linear,
    quadratic_matrix=basis_matrix_quadratic,
    priors=priors,
    roq_params=params,
    roq_scale_factor=scale_factor,
)

# write the weights to file so they can be loaded multiple times
likelihood.save_weights("weights.npz")

# remove the basis matrices as these are big for longer bases
del basis_matrix_linear, basis_matrix_quadratic

# load the weights from the file
likelihood = bilby.gw.likelihood.ROQGravitationalWaveTransient(
    interferometers=ifos,
    waveform_generator=search_waveform_generator,
    weights="weights.npz",
    priors=priors,
)

result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    npoints=500,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

# Resample the result using the full waveform model with the FakeSampler.
# This will give us an idea of how good a job the ROQ does.
full_likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)
resampled_result = bilby.run_sampler(
    likelihood=full_likelihood,
    priors=priors,
    sampler="fake_sampler",
    label="roq_resampled",
    outdir=outdir,
)

# Make a comparison corner plot with the two likelihoods.
bilby.core.result.plot_multiple([result, resampled_result], labels=["ROQ", "Regular"])



#!/bin/python
"""
Tutorial to demonstrate a new interferometer

We place a new instrument in Gingin, with an A+ sensitivity in a network of A+
interferometers at Hanford and Livingston.

This requires :code:`gwinc` to be installed. This is available via conda-forge.
"""

import bilby
import gwinc
import numpy as np

# Set the duration and sampling frequency of the data segment that we're going
# to inject the signal into
duration = 4
sampling_frequency = 1024

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "australian_detector"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(88170232)

# create a new detector using a PyGwinc sensitivity curve
curve = gwinc.load_budget("Aplus").run()

# Set up the detector as a four-kilometer detector in Gingin
# The location of this detector is not defined in Bilby, so we need to add it
AusIFO = bilby.gw.detector.Interferometer(
    power_spectral_density=bilby.gw.detector.PowerSpectralDensity(
        frequency_array=curve.freq, psd_array=curve.psd
    ),
    name="AusIFO",
    length=4,
    minimum_frequency=20,
    maximum_frequency=sampling_frequency / 2,
    latitude=-31.34,
    longitude=115.91,
    elevation=0.0,
    xarm_azimuth=2.0,
    yarm_azimuth=125.0,
)

# Set up two other detectors at Hanford and Livingston
interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
for ifo in interferometers:
    ifo.power_spectral_density = bilby.gw.detector.PowerSpectralDensity(
        frequency_array=curve.freq, psd_array=curve.psd
    )

# append the Australian detector to the list of other detectors
interferometers.append(AusIFO)


# Inject a gravitational-wave signal into the network
# as we're using a three-detector network of A+, we inject a GW150914-like
# signal at 4 Gpc
injection_parameters = dict(
    mass_1=36.0,
    mass_2=29.0,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=4000.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=0.2108,
)


# Fixed arguments passed into the source model
waveform_arguments = dict(waveform_approximant="IMRPhenomXP", reference_frequency=50.0)

# Create the waveform_generator using a LAL BinaryBlackHole source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    waveform_arguments=waveform_arguments,
)

start_time = injection_parameters["geocent_time"] + 2 - duration

# inject the signal into the interferometers

for ifo in interferometers:
    ifo.set_strain_data_from_power_spectral_density(
        sampling_frequency=sampling_frequency, duration=duration
    )
    ifo.inject_signal(
        parameters=injection_parameters, waveform_generator=waveform_generator
    )

    # plot the data for sanity
    signal = ifo.get_detector_response(
        waveform_generator.frequency_domain_strain(), injection_parameters
    )
    ifo.plot_data(signal=signal, outdir=outdir, label=label)

# set up priors
priors = bilby.gw.prior.BBHPriorDict()
for key in [
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "phi_12",
    "phi_jl",
    "psi",
    "geocent_time",
    "phase",
    "ra",
    "dec",
    "luminosity_distance",
    "theta_jn",
]:
    priors[key] = injection_parameters[key]

# Initialise the likelihood by passing in the interferometer data (IFOs)
# and the waveoform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=interferometers,
    waveform_generator=waveform_generator,
)


result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    npoints=1000,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

# make some plots of the outputs
result.plot_corner()
